/*
    Z80 Assembly IDE - An IDE for the Z80 line of TI calculators
    Copyright (C) 2006 Donald Straney

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <unistd.h>
#include <string.h>
#include <stdlib.h>
#include <gtk/gtk.h>
#include <gdk/gdkkeysyms.h>

#include "asmide.h"
#include "callback.h"
#include "settings.h"
#include "file.h"

void callback_text_deleted(GtkTextBuffer*, GtkTextIter*, GtkTextIter*, gpointer);
void callback_text_deleted_before(GtkTextBuffer*, GtkTextIter*, GtkTextIter*, gpointer);
void callback_text_inserted(GtkTextBuffer*, GtkTextIter*, gchar*, gint, gpointer);
void callback_modified_changed(GtkTextBuffer*, gpointer);
void callback_cursor_moved(GtkTextBuffer*, GtkTextIter*, GtkTextMark*, gpointer);
void callback_switch_file(GtkNotebook*, GtkNotebookPage*, guint, gpointer);
void callback_view_widget(GtkCheckMenuItem*, gpointer);
void callback_view_refresh(GtkWidget*, gpointer);
gboolean callback_closewindow(GtkWidget*, gpointer);
void callback_quit(GtkWidget*, gpointer);
gboolean callback_idle_colorall(gpointer);
//void callback_text_changed(GtkTextBuffer*, gpointer);
void update_line_color(GtkTextBuffer*, GtkTextIter);
void refresh_tabs_openfile(struct openfile *);
void add_to_undo_stack(int, GtkTextIter*, gchar*);
void callback_action_undo(GtkWidget*, gpointer);
void callback_action_redo(GtkWidget*, gpointer);
void update_undo_buttons();
void start_syncolor(struct openfile*);


GList *open_files = NULL;	//list of files that are open (pointers to openfile structs)
GtkWidget *output;		//text box with assembler output
GtkWidget *scrolledoutput;	//scrolled window with output text box
GtkWidget *statusbar;		//status bar
GtkWidget *vpaned;		//paned window dividing text box and assembler output
GtkWidget *toolbar;		//toolbar at top of window
GtkWidget *outputmenuitem;	//check menu item for 'show assembler output'
GList *file_widgets = NULL;	//list of widgets that are set insensitive unless a file is open
GtkTextIter *coloriter[NUM_SIMULT_FILES];	//text iter for background syntax-coloring when file is opened
guint statusbar_context;	//context ID for statusbar messages
struct openfile *curr_file = NULL;	//pointer to openfile struct of file that's currently open (and on top)
GtkWidget *window;		//toplevel window
gchar *color_label = NULL, *color_comment = NULL, *color_directive = NULL;	//names of colors used for syntax highlighting
gchar *output_type = NULL;	//type of output file (extension)
gchar *incdir1 = NULL, *incdir2 = NULL, *incdir3 = NULL;	//include directories
gboolean usesyncolor;		//use syntax coloring or not?
gboolean usewordwrap;		//use word wrap or not?
gboolean useautoindent;		//use auto indent or not?
gint tabwidth;			//width of tabs in chars
struct status *curr_status;	//current status of program
gunichar uchar_colon, uchar_uscore, uchar_hash, uchar_dot;	//(speeds up the coloring routine)
GtkWidget *notebook;		//pages used to switch between files
PangoFontDescription *font;	//editor font
GtkTextTagTable *tagtable;	//table with all text tags
GtkTextTag *tag_label, *tag_comment, *tag_directive;	//text tags for syntax coloring
GtkWidget *undo_widget_1, *undo_widget_2, *redo_widget_1, *redo_widget_2;	//undo/redo command widgets
gboolean save_actions;		//controls whether buffer modification callbacks for undo/redo are active


int main(int argc, char **argv) {

    gtk_init(&argc, &argv);

    if (argc == 2 && (!strcmp("--help", argv[1]) || !strcmp("-h", argv[1]))) {	//parse command line args
		printf("Z80 Assembly IDE\nCopyright Donald Straney 2006\n\nUsage: z80asmide [-h/--help | (filename) | (filename1) (filename2)... ]\n\n-h, --help: Shows this help\n(filename): Opens file in editor\n(Make sure any spaces in the file name are escaped)\n(filename1) (filename2)... : Opens files in editor\n");
		return 0;
    }

    uchar_colon = g_utf8_get_char(":");			//init unicode constants
    uchar_uscore = g_utf8_get_char("_");
    uchar_hash = g_utf8_get_char("#");
    uchar_dot = g_utf8_get_char(".");

	save_actions = TRUE;
    gtk_window_set_default_icon_name("z80asmide");

    window = gtk_window_new(GTK_WINDOW_TOPLEVEL);		//create the toplevel window
    g_signal_connect(window, "destroy", gtk_main_quit, NULL);
    g_signal_connect(window, "delete_event", callback_closewindow, NULL);
    gtk_container_set_border_width(window, 0);			//and set its properties

	notebook = gtk_notebook_new();
	g_signal_connect(notebook, "switch-page", callback_switch_file, NULL);

    output = gtk_text_view_new();
    scrolledoutput = gtk_scrolled_window_new(NULL, NULL);
    gtk_container_add(scrolledoutput, output);
    gtk_scrolled_window_set_shadow_type(scrolledoutput, GTK_SHADOW_IN);
    gtk_text_view_set_editable(output, FALSE);
    gtk_text_view_set_cursor_visible(output, FALSE);
    gtk_scrolled_window_set_policy(scrolledoutput, GTK_POLICY_AUTOMATIC, GTK_POLICY_AUTOMATIC);

    vpaned = gtk_vpaned_new();				//create the paned window for dividing text/output
    gtk_paned_pack1(vpaned, notebook, TRUE, FALSE);		//add the text box
    gtk_paned_pack2(vpaned, scrolledoutput, TRUE, FALSE);		//and output window to it

    statusbar = gtk_statusbar_new();			//create the status bar
    statusbar_context = gtk_statusbar_get_context_id(statusbar, "position");

    curr_status = (struct status *)g_malloc(sizeof(struct status));	//create a status struct to pass args to functions
    curr_status->window = window;					//populate it with data
    curr_status->color_label = &color_label;
    curr_status->color_comment = &color_comment;
    curr_status->color_directive = &color_directive;
    curr_status->output_type = &output_type;
    curr_status->curr_file = &curr_file;
    curr_status->usesyncolor = &usesyncolor;
    curr_status->usewordwrap = &usewordwrap;
    curr_status->useautoindent = &useautoindent;
    curr_status->tabwidth = &tabwidth;
	curr_status->font = &font;
	curr_status->notebook = notebook;
    curr_status->incdir1 = &incdir1;
    curr_status->incdir2 = &incdir2;
    curr_status->incdir3 = &incdir3;

    toolbar = gtk_toolbar_new();					//create a toolbar
    GtkWidget *toolbutton = gtk_tool_button_new_from_stock(GTK_STOCK_NEW);
    g_signal_connect(toolbutton, "clicked", callback_action_new, NULL);
    gtk_toolbar_insert(toolbar, toolbutton, -1);
    toolbutton = gtk_tool_button_new_from_stock(GTK_STOCK_OPEN);
    g_signal_connect(toolbutton, "clicked", callback_action_open, curr_status);
    gtk_toolbar_insert(toolbar, toolbutton, -1);
    toolbutton = gtk_tool_button_new_from_stock(GTK_STOCK_SAVE);
    file_widgets = g_list_prepend(file_widgets, toolbutton);
    g_signal_connect(toolbutton, "clicked", callback_action_save, curr_status);
    gtk_toolbar_insert(toolbar, toolbutton, -1);
    toolbutton = gtk_tool_button_new_from_stock(GTK_STOCK_CLOSE);
    file_widgets = g_list_prepend(file_widgets, toolbutton);
    g_signal_connect(toolbutton, "clicked", callback_action_close, curr_status);
    gtk_toolbar_insert(toolbar, toolbutton, -1);
    toolbutton = gtk_separator_tool_item_new();
    gtk_toolbar_insert(toolbar, toolbutton, -1);
    toolbutton = gtk_tool_button_new_from_stock(GTK_STOCK_UNDO);
    file_widgets = g_list_prepend(file_widgets, toolbutton);
	undo_widget_1 = toolbutton;
	g_signal_connect(toolbutton, "clicked", callback_action_undo, NULL);
    gtk_toolbar_insert(toolbar, toolbutton, -1);
    toolbutton = gtk_tool_button_new_from_stock(GTK_STOCK_REDO);
    file_widgets = g_list_prepend(file_widgets, toolbutton);
	redo_widget_1 = toolbutton;
	g_signal_connect(toolbutton, "clicked", callback_action_redo, NULL);
    gtk_toolbar_insert(toolbar, toolbutton, -1);
    toolbutton = gtk_tool_button_new_from_stock(GTK_STOCK_COPY);
    file_widgets = g_list_prepend(file_widgets, toolbutton);
    g_signal_connect(toolbutton, "clicked", callback_action_copy, curr_status);
    gtk_toolbar_insert(toolbar, toolbutton, -1);
    toolbutton = gtk_tool_button_new_from_stock(GTK_STOCK_CUT);
    file_widgets = g_list_prepend(file_widgets, toolbutton);
    g_signal_connect(toolbutton, "clicked", callback_action_cut, curr_status);
    gtk_toolbar_insert(toolbar, toolbutton, -1);
    toolbutton = gtk_tool_button_new_from_stock(GTK_STOCK_PASTE);
    file_widgets = g_list_prepend(file_widgets, toolbutton);
    g_signal_connect(toolbutton, "clicked", callback_action_paste, curr_status);
    gtk_toolbar_insert(toolbar, toolbutton, -1);
    toolbutton = gtk_tool_button_new_from_stock(GTK_STOCK_FIND);
    g_signal_connect(toolbutton, "clicked", callback_action_find, curr_status);
    file_widgets = g_list_prepend(file_widgets, toolbutton);
    gtk_toolbar_insert(toolbar, toolbutton, -1);
    toolbutton = gtk_separator_tool_item_new();
    gtk_toolbar_insert(toolbar, toolbutton, -1);
    toolbutton = gtk_tool_button_new(gtk_image_new_from_stock(GTK_STOCK_EXECUTE, GTK_ICON_SIZE_SMALL_TOOLBAR), "Build");
    file_widgets = g_list_prepend(file_widgets, toolbutton);
    g_signal_connect(toolbutton, "clicked", callback_action_build, curr_status);
    gtk_toolbar_insert(toolbar, toolbutton, -1);
    toolbutton = gtk_tool_button_new(gtk_image_new_from_stock(GTK_STOCK_PROPERTIES, GTK_ICON_SIZE_SMALL_TOOLBAR), "Code Counter");
    file_widgets = g_list_prepend(file_widgets, toolbutton);
	g_signal_connect(toolbutton, "clicked", callback_action_countcode, curr_status);
    gtk_toolbar_insert(toolbar, toolbutton, -1);
    gtk_toolbar_set_style(toolbar, GTK_TOOLBAR_ICONS);
    gtk_toolbar_set_icon_size(toolbar, GTK_ICON_SIZE_SMALL_TOOLBAR);
    //TODO: add toolbutton for "view output window"?

    GtkAccelGroup *accelgrp = gtk_accel_group_new();		//create an accelerator group to go with it
    gtk_window_add_accel_group(window, accelgrp);
    
    GtkWidget *menubar = gtk_menu_bar_new();			//now create the menu bar for the top of the window

    GtkWidget *menu = gtk_menu_new();				//and build the menus
    gtk_menu_set_accel_group(menu, accelgrp);
    gtk_menu_set_accel_path(menu, "<Z80AsmIDE-Document>/File");
    GtkWidget *menuitem = gtk_image_menu_item_new_from_stock(GTK_STOCK_NEW, accelgrp);
    g_signal_connect(menuitem, "activate", callback_action_new, NULL);
    gtk_menu_shell_append(menu, menuitem);
    menuitem = gtk_image_menu_item_new_from_stock(GTK_STOCK_OPEN, accelgrp);
    g_signal_connect(menuitem, "activate", callback_action_open, curr_status);
    gtk_menu_shell_append(menu, menuitem);
    menuitem = gtk_image_menu_item_new_from_stock(GTK_STOCK_SAVE, accelgrp);
    file_widgets = g_list_prepend(file_widgets, menuitem);
    g_signal_connect(menuitem, "activate", callback_action_save, curr_status);
    gtk_menu_shell_append(menu, menuitem);
    menuitem = gtk_image_menu_item_new_from_stock(GTK_STOCK_SAVE_AS, accelgrp);
    file_widgets = g_list_prepend(file_widgets, menuitem);
    g_signal_connect(menuitem, "activate", callback_action_saveas, curr_status);
    gtk_menu_shell_append(menu, menuitem);
    menuitem = gtk_image_menu_item_new_from_stock(GTK_STOCK_REVERT_TO_SAVED, accelgrp);
    file_widgets = g_list_prepend(file_widgets, menuitem);
    g_signal_connect(menuitem, "activate", callback_action_revert, curr_status);
    gtk_menu_shell_append(menu, menuitem);
    menuitem = gtk_image_menu_item_new_from_stock(GTK_STOCK_CLOSE, accelgrp);
    file_widgets = g_list_prepend(file_widgets, menuitem);
    g_signal_connect(menuitem, "activate", callback_action_close, curr_status);
    gtk_menu_shell_append(menu, menuitem);
/*    menuitem = gtk_separator_menu_item_new();
    gtk_menu_shell_append(menu, menuitem);
    menuitem = gtk_image_menu_item_new_from_stock(GTK_STOCK_PRINT, accelgrp);
    file_widgets = g_list_prepend(file_widgets, menuitem);
    gtk_menu_shell_append(menu, menuitem);*/
    menuitem = gtk_separator_menu_item_new();
    gtk_menu_shell_append(menu, menuitem);
    menuitem = gtk_image_menu_item_new_from_stock(GTK_STOCK_QUIT, accelgrp);
    g_signal_connect(menuitem, "activate", callback_quit, NULL);
    gtk_menu_shell_append(menu, menuitem);
    menuitem = gtk_menu_item_new_with_mnemonic("_File");
    gtk_menu_item_set_submenu(menuitem, menu);
    gtk_menu_shell_append(menubar, menuitem);

    menu = gtk_menu_new();
    gtk_menu_set_accel_group(menu, accelgrp);
    gtk_menu_set_accel_path(menu, "<Z80AsmIDE-Document>/Edit");
    menuitem = gtk_image_menu_item_new_from_stock(GTK_STOCK_UNDO, accelgrp);
	undo_widget_2 = menuitem;
	g_signal_connect(menuitem, "activate", callback_action_undo, NULL);
    gtk_menu_shell_append(menu, menuitem);
    menuitem = gtk_image_menu_item_new_from_stock(GTK_STOCK_REDO, accelgrp);
	redo_widget_2 = menuitem;
	g_signal_connect(menuitem, "activate", callback_action_redo, NULL);
    gtk_menu_shell_append(menu, menuitem);
    menuitem = gtk_separator_menu_item_new();
    gtk_menu_shell_append(menu, menuitem);
    menuitem = gtk_image_menu_item_new_from_stock(GTK_STOCK_COPY, accelgrp);
    g_signal_connect(menuitem, "activate", callback_action_copy, curr_status);
    gtk_menu_shell_append(menu, menuitem);
    menuitem = gtk_image_menu_item_new_from_stock(GTK_STOCK_CUT, accelgrp);
    g_signal_connect(menuitem, "activate", callback_action_cut, curr_status);
    gtk_menu_shell_append(menu, menuitem);
    menuitem = gtk_image_menu_item_new_from_stock(GTK_STOCK_PASTE, accelgrp);
    g_signal_connect(menuitem, "activate", callback_action_paste, curr_status);
    gtk_menu_shell_append(menu, menuitem);
    menuitem = gtk_image_menu_item_new_with_mnemonic("Select _All");
    gtk_image_menu_item_set_image(menuitem, gtk_image_new_from_stock(GTK_STOCK_JUSTIFY_CENTER, GTK_ICON_SIZE_MENU));
    g_signal_connect(menuitem, "activate", callback_action_selectall, curr_status);
    gtk_menu_shell_append(menu, menuitem);
    menuitem = gtk_separator_menu_item_new();
    gtk_menu_shell_append(menu, menuitem);
    menuitem = gtk_image_menu_item_new_from_stock(GTK_STOCK_FIND, accelgrp);
    g_signal_connect(menuitem, "activate", callback_action_find, curr_status);
    gtk_menu_shell_append(menu, menuitem);
    menuitem = gtk_image_menu_item_new_from_stock(GTK_STOCK_FIND_AND_REPLACE, accelgrp);
    g_signal_connect(menuitem, "activate", callback_action_replace, curr_status);
    gtk_menu_shell_append(menu, menuitem);
    menuitem = gtk_image_menu_item_new_with_mnemonic("_Go To Line");
    gtk_image_menu_item_set_image(menuitem, gtk_image_new_from_stock(GTK_STOCK_JUMP_TO, GTK_ICON_SIZE_MENU));
    g_signal_connect(menuitem, "activate", callback_action_gotoline, curr_status);
    gtk_menu_shell_append(menu, menuitem);
    menuitem = gtk_menu_item_new_with_mnemonic("Go To _Label");
    g_signal_connect(menuitem, "activate", callback_action_gotolabel, curr_status);
    gtk_menu_shell_append(menu, menuitem);
    menuitem = gtk_menu_item_new_with_mnemonic("_Edit");
    file_widgets = g_list_prepend(file_widgets, menuitem);
    gtk_menu_item_set_submenu(menuitem, menu);
    gtk_menu_shell_append(menubar, menuitem);

    menu = gtk_menu_new();
    gtk_menu_set_accel_group(menu, accelgrp);
    gtk_menu_set_accel_path(menu, "<Z80AsmIDE-Document>/Settings");
    menuitem = gtk_image_menu_item_new_with_mnemonic("_Editor");
    g_signal_connect(menuitem, "activate", callback_action_seteditoropt, curr_status);
    gtk_image_menu_item_set_image(menuitem, gtk_image_new_from_stock(GTK_STOCK_EDIT, GTK_ICON_SIZE_MENU));
    gtk_menu_shell_append(menu, menuitem);
    menuitem = gtk_image_menu_item_new_with_mnemonic("_Assembler");
    g_signal_connect(menuitem, "activate", callback_action_setassembleropt, curr_status);
    gtk_image_menu_item_set_image(menuitem, gtk_image_new_from_stock(GTK_STOCK_PROPERTIES, GTK_ICON_SIZE_MENU));
    gtk_menu_shell_append(menu, menuitem);
    menuitem = gtk_image_menu_item_new_with_mnemonic("_Output Format");
    gtk_image_menu_item_set_image(menuitem, gtk_image_new_from_stock(GTK_STOCK_HARDDISK, GTK_ICON_SIZE_MENU));
    g_signal_connect(menuitem, "activate", callback_action_setoutputtype, curr_status);
    gtk_menu_shell_append(menu, menuitem);
    menuitem = gtk_menu_item_new_with_mnemonic("_Settings");
    gtk_menu_item_set_submenu(menuitem, menu);
    gtk_menu_shell_append(menubar, menuitem);

    menu = gtk_menu_new();
    gtk_menu_set_accel_group(menu, accelgrp);
    gtk_menu_set_accel_path(menu, "<Z80AsmIDE-Document>/View");
    menuitem = gtk_check_menu_item_new_with_mnemonic("Assembler _Output");
    outputmenuitem = menuitem;
    g_signal_connect(menuitem, "activate", callback_view_widget, scrolledoutput);
    gtk_menu_shell_append(menu, menuitem);
    menuitem = gtk_check_menu_item_new_with_mnemonic("_Toolbar");
    gtk_check_menu_item_set_active(menuitem, TRUE);
    g_signal_connect(menuitem, "activate", callback_view_widget, toolbar);
    gtk_menu_shell_append(menu, menuitem);
    menuitem = gtk_check_menu_item_new_with_mnemonic("_Status Bar");
    gtk_check_menu_item_set_active(menuitem, TRUE);
    g_signal_connect(menuitem, "activate", callback_view_widget, statusbar);
    gtk_menu_shell_append(menu, menuitem);
    menuitem = gtk_menu_item_new_with_mnemonic("_View");
    file_widgets = g_list_prepend(file_widgets, menuitem);
    gtk_menu_item_set_submenu(menuitem, menu);
    gtk_menu_shell_append(menubar, menuitem);

    menu = gtk_menu_new();
    gtk_menu_set_accel_group(menu, accelgrp);
    gtk_menu_set_accel_path(menu, "<Z80AsmIDE-Document>/Assemble");
    menuitem = gtk_image_menu_item_new_with_mnemonic("_Build Program");
    gtk_image_menu_item_set_image(menuitem, gtk_image_new_from_stock(GTK_STOCK_EXECUTE, GTK_ICON_SIZE_MENU));
    g_signal_connect(menuitem, "activate", callback_action_build, curr_status);
    gtk_menu_shell_append(menu, menuitem);
    menuitem = gtk_image_menu_item_new_with_mnemonic("Create Symbol _Table");
    gtk_image_menu_item_set_image(menuitem, gtk_image_new_from_stock(GTK_STOCK_INDEX, GTK_ICON_SIZE_MENU));
    g_signal_connect(menuitem, "activate", callback_action_createsymtable, curr_status);
    gtk_menu_shell_append(menu, menuitem);
    menuitem = gtk_image_menu_item_new_with_mnemonic("_Code Counter");
    gtk_image_menu_item_set_image(menuitem, gtk_image_new_from_stock(GTK_STOCK_PROPERTIES, GTK_ICON_SIZE_MENU));
    g_signal_connect(menuitem, "activate", callback_action_countcode, curr_status);
    gtk_menu_shell_append(menu, menuitem);
    menuitem = gtk_image_menu_item_new_with_mnemonic("Program _Statistics");
    gtk_image_menu_item_set_image(menuitem, gtk_image_new_from_stock(GTK_STOCK_INFO, GTK_ICON_SIZE_MENU));
    g_signal_connect(menuitem, "activate", callback_action_printstats, curr_status);
    gtk_menu_shell_append(menu, menuitem);
    menuitem = gtk_image_menu_item_new_with_mnemonic("Generate _Listing");
    gtk_image_menu_item_set_image(menuitem, gtk_image_new_from_stock(GTK_STOCK_INDEX, GTK_ICON_SIZE_MENU));
    g_signal_connect(menuitem, "activate", callback_action_createlisting, curr_status);
    gtk_menu_shell_append(menu, menuitem);
    menuitem = gtk_menu_item_new_with_mnemonic("_Assemble");
    file_widgets = g_list_prepend(file_widgets, menuitem);
    gtk_menu_item_set_submenu(menuitem, menu);
    gtk_menu_shell_append(menubar, menuitem);

    menu = gtk_menu_new();
    gtk_menu_set_accel_group(menu, accelgrp);
    gtk_menu_set_accel_path(menu, "<Z80AsmIDE-Document>/Help");
    menuitem = gtk_image_menu_item_new_with_mnemonic("_Manual");
    gtk_image_menu_item_set_image(menuitem, gtk_image_new_from_stock(GTK_STOCK_HELP, GTK_ICON_SIZE_MENU));
    g_signal_connect(menuitem, "activate", callback_action_showhelp, NULL);
    gtk_menu_shell_append(menu, menuitem);
    menuitem = gtk_image_menu_item_new_from_stock(GTK_STOCK_ABOUT, accelgrp);
    g_signal_connect(menuitem, "activate", callback_action_about, NULL);
    gtk_menu_shell_append(menu, menuitem);
    menuitem = gtk_menu_item_new_with_mnemonic("_Help");
    gtk_menu_item_set_submenu(menuitem, menu);
    gtk_menu_shell_append(menubar, menuitem);

    gtk_accel_map_add_entry("<Z80AsmIDE-Document>/Edit/Select All", GDK_A, GDK_CONTROL_MASK);	//set up accelerators
    gtk_accel_map_add_entry("<Z80AsmIDE-Document>/Edit/Go To Line", GDK_G, GDK_CONTROL_MASK);
    gtk_accel_map_add_entry("<Z80AsmIDE-Document>/Edit/Go To Label", GDK_L, GDK_CONTROL_MASK);
    gtk_accel_map_add_entry("<Z80AsmIDE-Document>/Edit/Undo", GDK_Z, GDK_CONTROL_MASK);
    gtk_accel_map_add_entry("<Z80AsmIDE-Document>/Edit/Redo", GDK_Z, GDK_CONTROL_MASK | GDK_SHIFT_MASK);
    gtk_accel_map_add_entry("<Z80AsmIDE-Document>/Assemble/Build Program", GDK_F5, 0);
    gtk_accel_map_add_entry("<Z80AsmIDE-Document>/Assemble/Create Symbol Table", GDK_F6, 0);
    gtk_accel_map_add_entry("<Z80AsmIDE-Document>/Assemble/Code Counter", GDK_F7, 0);
    gtk_accel_map_add_entry("<Z80AsmIDE-Document>/Assemble/Program Statistics", GDK_F8, 0);
    gtk_accel_map_add_entry("<Z80AsmIDE-Document>/Assemble/Generate Listing", GDK_F9, 0);
    gtk_accel_map_add_entry("<Z80AsmIDE-Document>/Help/Using Assembly IDE", GDK_F1, 0);

    GtkWidget *vbox = gtk_vbox_new(FALSE, 0);			//create a vertical box to hold everything
    gtk_container_add(window, vbox);
    gtk_box_pack_start(vbox, menubar, FALSE, FALSE, 0);		//and shove it all in!
    gtk_box_pack_start(vbox, toolbar, FALSE, FALSE, 0);
    gtk_box_pack_start(vbox, vpaned, TRUE, TRUE, 2);
    gtk_box_pack_start(vbox, statusbar, FALSE, FALSE, 0);

    load_settings(curr_status);	//load settings

    tagtable = gtk_text_tag_table_new();		//create a new tag table for the color tags
    tag_comment = gtk_text_tag_new("comment");		//create the tags
    tag_label = gtk_text_tag_new("label");
    tag_directive = gtk_text_tag_new("directive");
    gtk_text_tag_table_add(tagtable, tag_comment);		//and add them to the table
    gtk_text_tag_table_add(tagtable, tag_label);
    gtk_text_tag_table_add(tagtable, tag_directive);
    refresh_tags();

    gint window_height;
    gtk_window_get_size(window, NULL, &window_height);
    gtk_paned_set_position(vpaned, window_height - DEFAULT_PANEDPOS);

    gtk_widget_show_all(window);
    gtk_widget_hide(scrolledoutput);
    update_titlebar();					//update the titlebar text
    update_view();					//and current view

    int i;
    for (i = 0; i < NUM_SIMULT_FILES; i++)
		coloriter[i] = NULL;

    if (argc > 1) {			//if the command line options specified a file or list of files to load,

		if (argc > 1 + NUM_SIMULT_FILES) {
	    	//TODO: should it also show a warning dialog?
	    	fprintf(stderr, "z80asmide: Can't open more than %d files at a time,\nonly opening the first %d\n", NUM_SIMULT_FILES, NUM_SIMULT_FILES);
	    	argc = 1 + NUM_SIMULT_FILES;
		}

		for (; argc > 1; argc--) {
    	    gchar *filecontents;
	    	guint filelength;
	    	if (open_file_backend(argv[argc - 1], &filecontents, &filelength)) {	//then open them
				add_file(argv[argc - 1], filecontents, filelength);
				g_free(filecontents);
	    	}
		}
    }

    gtk_main();
}


/*
    Returns allocated text for
    file [file]
*/

gchar *get_file_text(struct openfile *file) {

    GtkTextBuffer *textbuffer = file->textbuffer;
    GtkTextIter startiter, enditer;
    gtk_text_buffer_get_bounds(textbuffer, &startiter, &enditer);
    return gtk_text_buffer_get_text(textbuffer, &startiter, &enditer, TRUE);
}


/*
    Re-sets the tab stops
    for [file]
*/

void refresh_tabs_file(struct openfile *file) {

    const gchar *lang = g_getenv("LANG");	//try to get the current language

    if (lang == NULL)			//if it isn't available, use US english by default
		lang = "en_US";

    PangoFontMetrics *fontmetrics = pango_context_get_metrics(gtk_widget_get_pango_context(output), gtk_widget_get_style(output)->font_desc, pango_language_from_string(lang));
    int char_width = PANGO_PIXELS(pango_font_metrics_get_approximate_digit_width(fontmetrics));
    pango_font_metrics_unref(fontmetrics);			//get width of chars with current font for future reference

    gint set_tab_width = tabwidth * char_width;

    //TODO: do this with a loop or something
    gtk_text_view_set_tabs(file->textview, pango_tab_array_new_with_positions(20, TRUE, PANGO_TAB_LEFT, set_tab_width, PANGO_TAB_LEFT, set_tab_width * 2, PANGO_TAB_LEFT, set_tab_width * 3, PANGO_TAB_LEFT, set_tab_width * 4, PANGO_TAB_LEFT, set_tab_width * 5, PANGO_TAB_LEFT, set_tab_width * 6, PANGO_TAB_LEFT, set_tab_width * 7, PANGO_TAB_LEFT, set_tab_width * 8, PANGO_TAB_LEFT, set_tab_width * 9, PANGO_TAB_LEFT, set_tab_width * 10, PANGO_TAB_LEFT, set_tab_width * 11, PANGO_TAB_LEFT, set_tab_width * 12, PANGO_TAB_LEFT, set_tab_width * 13, PANGO_TAB_LEFT, set_tab_width * 14, PANGO_TAB_LEFT, set_tab_width * 15, PANGO_TAB_LEFT, set_tab_width * 16, PANGO_TAB_LEFT, set_tab_width * 17, PANGO_TAB_LEFT, set_tab_width * 18, PANGO_TAB_LEFT, set_tab_width * 19, PANGO_TAB_LEFT, set_tab_width * 20));
}


/*
    Adds file to list of open
    files, with name of [filename],
    contents in [filecontents], length
    of [filelength]
*/

void add_file(gchar *filename, gchar *filecontents, gint filelength) {

    struct openfile *newfile = g_malloc(sizeof(struct openfile));	//create some memory for the file's data

    strcpy(newfile->filename, filename);		//set its initial properties

    gchar *tablabel;
    if (strlen(filename) == 0)
		tablabel = g_strdup("(Untitled)");
    else
		tablabel = g_path_get_basename(filename);	//and a menu item for the "open files" menu

    GtkTextBuffer *textbuffer = gtk_text_buffer_new(tagtable);
    GtkWidget *textview = gtk_text_view_new_with_buffer(textbuffer);	//create the text box for editing
	gtk_widget_modify_font(textview, font);
    gtk_text_view_set_wrap_mode(textview, usewordwrap ? GTK_WRAP_WORD_CHAR : GTK_WRAP_NONE);
    GtkWidget *scrolledtext = gtk_scrolled_window_new(NULL, NULL);	//and a scrolled window for holding it
    gtk_container_set_border_width(scrolledtext, 1);		//TODO: is this needed?
    gtk_container_add(scrolledtext, textview);
    gtk_scrolled_window_set_shadow_type(scrolledtext, GTK_SHADOW_IN);
    gtk_scrolled_window_set_policy(scrolledtext, GTK_POLICY_AUTOMATIC, GTK_POLICY_AUTOMATIC);
	gtk_widget_show_all(scrolledtext);
	newfile->scrolledtext = scrolledtext;
	newfile->textview = textview;
	newfile->textbuffer = textbuffer;
	gint tab = gtk_notebook_append_page(notebook, scrolledtext, gtk_label_new(tablabel));
    g_free(tablabel);

    int i;
    for (i = 0; i < NUM_UNDO_LEVELS; i++)
		newfile->undo_stack[i] = NULL;

    newfile->undo_stack_pos = 0;

    gtk_text_buffer_set_text(textbuffer, filecontents, filelength);	//set its contents
    gtk_text_buffer_set_modified(textbuffer, FALSE);

//    g_signal_connect(textbuffer, "changed", callback_text_changed, NULL);	//and connect signals
    g_signal_connect(textbuffer, "mark-set", callback_cursor_moved, NULL);
    g_signal_connect(textbuffer, "modified-changed", callback_modified_changed, NULL);
    g_signal_connect_after(textbuffer, "delete-range", callback_text_deleted, NULL);
	g_signal_connect(textbuffer, "delete-range", callback_text_deleted_before, NULL);
    g_signal_connect_after(textbuffer, "insert-text", callback_text_inserted, NULL);

    curr_file = newfile;			//set it to be on top
    open_files = g_list_prepend(open_files, newfile);	//and add it to the list of open files

	gtk_notebook_set_current_page(notebook, tab);

    update_view();				//then update everything
//    update_titlebar();
	update_undo_buttons();
    refresh_tabs_file(newfile);
    init_file();
}


/*
    Removes current file
    from list of open files,
    returns TRUE if successful
*/

gboolean remove_file() {

    if (gtk_text_buffer_get_modified(curr_file->textbuffer)) {	//if the file has unsaved changes,
    
		gchar *filename;
		if (strlen(curr_file->filename) == 0)
	    	filename = strdup("(Untitled)");
		else
	    	filename = g_path_get_basename(curr_file->filename);
		GtkWidget *dialog = gtk_message_dialog_new(NULL, 0, GTK_MESSAGE_QUESTION, GTK_BUTTONS_NONE, "%s has unsaved changes", filename);
		gtk_message_dialog_format_secondary_text(dialog, "Do you want to save it?");
		gtk_dialog_add_buttons(dialog, GTK_STOCK_YES, GTK_RESPONSE_YES, GTK_STOCK_NO, GTK_RESPONSE_NO, GTK_STOCK_CANCEL, GTK_RESPONSE_CANCEL, NULL);
		gint result = gtk_dialog_run(dialog);			//ask the user if they want to save it
		gtk_widget_destroy(dialog);
		g_free(filename);

		if (result == GTK_RESPONSE_YES) {		//if yes was pressed, save file
	    	if (!save_file(curr_file, curr_status))		//(cancel if it didn't work)
				return FALSE;
 		} else if (result == GTK_RESPONSE_CANCEL) {
	    	return FALSE;				//don't exit if cancel was pressed!
		}
    }

    struct openfile *closefile = curr_file;

	int i;								//free the contents of the undo stack
	for (i = 0; i < NUM_UNDO_LEVELS; i++) {
		if (closefile->undo_stack[i] != NULL) {
			g_free(closefile->undo_stack[i]->text);
			g_free(closefile->undo_stack[i]);
		}
	}

    open_files = g_list_remove(open_files, curr_file);	//take the file out of the list
    
    if (g_list_length(open_files) == 0) {		//if that was the last open file,

		curr_file = NULL;			//set the current file pointer to null
		update_view();			//update everything to show the changes
		update_titlebar();

    } else {

		curr_file = (struct openfile *)g_list_nth_data(open_files, 0);	//otherwise, put one of the remaining files "on top"
		gtk_notebook_set_current_page(notebook, gtk_notebook_page_num(notebook, curr_file->scrolledtext));
		gtk_widget_grab_focus(curr_file->textview);
    }

	gtk_notebook_remove_page(notebook, gtk_notebook_page_num (notebook, closefile->scrolledtext));	//remove the tab
    g_free(closefile);				//then the whole struct

    return TRUE;
}


/*
    Closes all files,
    returns TRUE if
    successful
*/

gboolean remove_file_all() {

    gint i;
    for (i = g_list_length(open_files); i > 0; i--) {	//get the number of open files,
		if (!remove_file())				//and try to close each one
	    	return FALSE;				//if there's a problem closing it, then abort
    }
	
    return TRUE;
}


/*
	Adds [action] to undo stack,
	w/[text] inserted/deleted
	at position of [iter]
*/

void add_to_undo_stack(int action, GtkTextIter *iter, gchar *text) {

	if (!save_actions)
		return;

	struct change **undo_stack = curr_file->undo_stack;
	int undo_stack_pos = curr_file->undo_stack_pos;
	int offset = (action == 0 ? gtk_text_iter_get_offset(iter) - strlen(text) : gtk_text_iter_get_offset(iter));

	if (undo_stack_pos == NUM_UNDO_LEVELS) {

		g_free(undo_stack[0]->text);
		g_free(undo_stack[0]);
		int i;
		for (i = 0; i < NUM_UNDO_LEVELS - 1; i++)
			undo_stack[i] = undo_stack[i + 1];

		undo_stack[undo_stack_pos - 1] = (struct change *)malloc(sizeof(struct change));
		undo_stack[undo_stack_pos - 1]->type = action;
		undo_stack[undo_stack_pos - 1]->offset = offset;
		undo_stack[undo_stack_pos - 1]->text = g_strdup(text);

	} else if (undo_stack[undo_stack_pos] == NULL) {

		undo_stack[undo_stack_pos] = (struct change *)malloc(sizeof(struct change));
		undo_stack[undo_stack_pos]->type = action;
		undo_stack[undo_stack_pos]->offset = offset;
		undo_stack[undo_stack_pos]->text = g_strdup(text);
		curr_file->undo_stack_pos++;

	} else {
//		undo_stack_pos++;

		if (undo_stack[undo_stack_pos] != NULL) {
			int i;
			for (i = undo_stack_pos; i < NUM_UNDO_LEVELS; i++) {
				if (undo_stack[i] != NULL) {
					g_free(undo_stack[i]->text);
					g_free(undo_stack[i]);
					undo_stack[i] = NULL;
				}
			}
		}

		undo_stack[undo_stack_pos] = (struct change *)malloc(sizeof(struct change));
		undo_stack[undo_stack_pos]->type = action;
		undo_stack[undo_stack_pos]->offset = offset;
		undo_stack[undo_stack_pos]->text = g_strdup(text);
		curr_file->undo_stack_pos++;
	}

	update_undo_buttons();
}


/*
	Undoes last action from
	undo stack, updates pointer
*/

void callback_action_undo(GtkWidget *widget, gpointer data) {

	curr_file->undo_stack_pos--;
	struct change *action = curr_file->undo_stack[curr_file->undo_stack_pos];
	save_actions = FALSE;

	if (action->type == 1) {
		GtkTextIter iter;
		gtk_text_buffer_get_iter_at_offset(curr_file->textbuffer, &iter, action->offset);
		gtk_text_buffer_insert(curr_file->textbuffer, &iter, action->text, -1);
	} else {
		GtkTextIter iter, iter2;
		gtk_text_buffer_get_iter_at_offset(curr_file->textbuffer, &iter, action->offset);
		gtk_text_buffer_get_iter_at_offset(curr_file->textbuffer, &iter2, action->offset + strlen(action->text));
		gtk_text_buffer_delete(curr_file->textbuffer, &iter, &iter2);
	}

	save_actions = TRUE;
	update_undo_buttons();
}


/*
	Redoes next action on
	undo stack, updates pointer
*/

void callback_action_redo(GtkWidget *widget, gpointer data) {

	save_actions = FALSE;
	struct change *action = curr_file->undo_stack[curr_file->undo_stack_pos];

	if (action->type == 0) {
		GtkTextIter iter;
		gtk_text_buffer_get_iter_at_offset(curr_file->textbuffer, &iter, action->offset);
		gtk_text_buffer_insert(curr_file->textbuffer, &iter, action->text, -1);
	} else {
		GtkTextIter iter, iter2;
		gtk_text_buffer_get_iter_at_offset(curr_file->textbuffer, &iter, action->offset);
		gtk_text_buffer_get_iter_at_offset(curr_file->textbuffer, &iter2, action->offset + strlen(action->text));
		gtk_text_buffer_delete(curr_file->textbuffer, &iter, &iter2);
	}

	curr_file->undo_stack_pos++;
	save_actions = TRUE;
	update_undo_buttons();
}


/*
    Closes all files and exits
    when window's close button
    is pressed
*/

gboolean callback_closewindow(GtkWidget *widget, gpointer data) {

    save_settings(curr_status);	//save settings

    if (remove_file_all()) {		//try to close all files,
		g_free(color_label);
		g_free(color_comment);
		g_free(color_directive);
		g_free(output_type);
		g_free(curr_status);
		g_free(incdir1);
		g_free(incdir2);
		g_free(incdir3);
		pango_font_description_free(font);
		return FALSE;			//and if that was successful, exit
    } else {
		return TRUE;
    }
}


/*
    Closes all files and exits
    when 'quit' menu item is activated
*/

void callback_quit(GtkWidget *widget, gpointer data) {

	if (!callback_closewindow(widget, data))
		gtk_main_quit();
}


/*
    Sets cursor to beginning of file,
    grabs focus, and updates colors
*/

void init_file() {

    GtkTextIter iter;
    gtk_text_buffer_get_start_iter(curr_file->textbuffer, &iter);	//set the cursor to the beginning of the file
    gtk_text_buffer_place_cursor(curr_file->textbuffer, &iter);

    if (usesyncolor)						//start the background function to add the syntax coloring
		start_syncolor(curr_file);

    gtk_widget_grab_focus(curr_file->textview);
	update_titlebar();
}


/*
    Starts syntax coloring for [file]
*/

void start_syncolor(struct openfile *file) {

    GtkTextIter iter;
    gtk_text_buffer_get_start_iter(file->textbuffer, &iter);

    int i;
    for (i = 0; i < NUM_SIMULT_FILES; i++) {
		if (coloriter[i] == NULL)
	    	break;
    }
    
    coloriter[i] = gtk_text_iter_copy(&iter);
    g_idle_add(callback_idle_colorall, (gpointer)i);
}


/*
    Sets/unsets syntax coloring
    on all open files
*/

void set_unset_syncolor() {

	GList *node;
	for (node = open_files; node != NULL; node = g_list_next(node)) {	//for each file,
		struct openfile *file = (struct openfile *)node->data;
		if (usesyncolor) {					//if syntax coloring is on,
	    	start_syncolor(file);				//then start the background syntax coloring function
		} else {
	    	GtkTextIter startiter, enditer;			//otherwise, remove all tags from the buffer
	    	gtk_text_buffer_get_bounds(file->textbuffer, &startiter, &enditer);
	    	gtk_text_buffer_remove_all_tags(file->textbuffer, &startiter, &enditer);
		}
	}
}


/*
    Moves file pointed to by
    [data] to top
*/

void callback_switch_file(GtkNotebook *notebook, GtkNotebookPage *page, guint page_num, gpointer data) {

	GList *node;
	for (node = open_files; node != NULL; node = g_list_next(node)) {
		struct openfile *file = (struct openfile *)node->data;
		gint tab = gtk_notebook_page_num(notebook, file->scrolledtext);
		if (tab == (gint)page_num) {
			curr_file = file;
			GtkTextIter iter;
			GtkTextMark *mark = gtk_text_buffer_get_insert(curr_file->textbuffer);
			gtk_text_buffer_get_iter_at_mark(curr_file->textbuffer, &iter, mark);
			callback_cursor_moved(curr_file->textbuffer, &iter, mark, NULL);
			update_view();
			update_titlebar();
			update_undo_buttons();
			gtk_widget_grab_focus(curr_file->textview);
			break;
		}
	}

	//TODO: show error message here
}


/*
	Sets undo/redo widgets
	sensitive or insensitive
	depending on the position
	in the undo stack
*/

void update_undo_buttons() {

	printf ("%d\n", curr_file->undo_stack_pos);

	if (curr_file->undo_stack_pos == 0) {
		gtk_widget_set_sensitive(undo_widget_1, FALSE);
		gtk_widget_set_sensitive(undo_widget_2, FALSE);
	} else {
		gtk_widget_set_sensitive(undo_widget_1, TRUE);
		gtk_widget_set_sensitive(undo_widget_2, TRUE);
	}

	if (curr_file->undo_stack_pos == NUM_UNDO_LEVELS || curr_file->undo_stack[curr_file->undo_stack_pos] == NULL) {
		gtk_widget_set_sensitive(redo_widget_1, FALSE);
		gtk_widget_set_sensitive(redo_widget_2, FALSE);
	} else {
		gtk_widget_set_sensitive(redo_widget_1, TRUE);
		gtk_widget_set_sensitive(redo_widget_2, TRUE);
	}
}


/*
    Updates contents of main
    text box for current
    file that's "on top"
*/

void update_view() {

    if (curr_file == NULL) {		//if no files are open,

		GList *list = file_widgets;	//set all the file-related widgets insensitive
		while (list != NULL) {
	    	gtk_widget_set_sensitive(GTK_WIDGET(list->data), FALSE);
	    	list = g_list_next(list);
		}
		
		gtk_widget_hide(vpaned);	//and hide the text box
		gtk_widget_hide(statusbar);

    } else {

		GList *list = file_widgets;	//otherwise, set all the file-related widgets sensitive,
		while (list != NULL) {
	    	gtk_widget_set_sensitive(GTK_WIDGET(list->data), TRUE);
	    	list = g_list_next(list);
		}

		gtk_widget_show(vpaned);	//make sure the text box is visible,
		gtk_widget_show(statusbar);
    }
}


/*
    Updates text in titlebar
    and tab label for current file
*/

void update_titlebar() {

    gchar *titletext;

    if (curr_file != NULL) {		//if there's a file open,
		gchar *savedtext, *filetext;

		if (!gtk_text_buffer_get_modified(curr_file->textbuffer))
	    	savedtext = "";
		else
	    	savedtext = "*";

		if (strlen(curr_file->filename) == 0)
	    	filetext = g_strdup("(Untitled)");
		else
	    	filetext = g_path_get_basename(curr_file->filename);

		titletext = g_strdup_printf("%s%s", savedtext, filetext);	//set the title with the filename
		gtk_notebook_set_tab_label_text(notebook, curr_file->scrolledtext, titletext);
		g_free(filetext);

    } else {

		titletext = g_strdup("Z80 Assembly IDE");	//otherwise just show the name of the program
    }
    
    gtk_window_set_title(window, titletext);
    g_free(titletext);
}


/*
    Toggles visibility of
    widget in [data]
*/

void callback_view_widget(GtkCheckMenuItem *widget, gpointer data) {

    if (widget->active)
		gtk_widget_show(GTK_WIDGET(data));
    else
		gtk_widget_hide(GTK_WIDGET(data));
}


/*
    Re-sets word wrap settings
    on all open buffers
*/

void refresh_wordwrap() {

    GList *node;
    for (node = open_files; node != NULL; node = g_list_next(node)) {
		struct openfile *file = (struct openfile *)node->data;
		gtk_text_view_set_wrap_mode(file->textview, usewordwrap ? GTK_WRAP_WORD_CHAR : GTK_WRAP_NONE);
    }
}


/*
    Re-sets color of color
    tags on all open buffers
*/

void refresh_tags() {

    g_object_set(tag_label, "foreground", color_label, NULL);
    g_object_set(tag_comment, "foreground", color_comment, NULL);
    g_object_set(tag_directive, "foreground", color_directive, NULL);
}


/*
    Re-sets tab stops
    for all files
*/

void refresh_tabs() {	//TODO: give this a better name, like refresh_font

    const gchar *lang = g_getenv("LANG");	//try to get the current language

    if (lang == NULL)			//if it isn't available, use US english by default
		lang = "en_US";

    PangoFontMetrics *fontmetrics = pango_context_get_metrics(gtk_widget_get_pango_context(output), gtk_widget_get_style(output)->font_desc, pango_language_from_string(lang));
    int char_width = PANGO_PIXELS(pango_font_metrics_get_approximate_digit_width(fontmetrics));
    pango_font_metrics_unref(fontmetrics);			//get width of chars with current font for future reference

    gint set_tab_width = tabwidth * char_width;
	GList *node;
	for (node = open_files; node != NULL; node = g_list_next(node)) {
		struct openfile *file = (struct openfile *)node->data;
	    //TODO: do this with a loop or something
		gtk_text_view_set_tabs(file->textview, pango_tab_array_new_with_positions(20, TRUE, PANGO_TAB_LEFT, set_tab_width, PANGO_TAB_LEFT, set_tab_width * 2, PANGO_TAB_LEFT, set_tab_width * 3, PANGO_TAB_LEFT, set_tab_width * 4, PANGO_TAB_LEFT, set_tab_width * 5, PANGO_TAB_LEFT, set_tab_width * 6, PANGO_TAB_LEFT, set_tab_width * 7, PANGO_TAB_LEFT, set_tab_width * 8, PANGO_TAB_LEFT, set_tab_width * 9, PANGO_TAB_LEFT, set_tab_width * 10, PANGO_TAB_LEFT, set_tab_width * 11, PANGO_TAB_LEFT, set_tab_width * 12, PANGO_TAB_LEFT, set_tab_width * 13, PANGO_TAB_LEFT, set_tab_width * 14, PANGO_TAB_LEFT, set_tab_width * 15, PANGO_TAB_LEFT, set_tab_width * 16, PANGO_TAB_LEFT, set_tab_width * 17, PANGO_TAB_LEFT, set_tab_width * 18, PANGO_TAB_LEFT, set_tab_width * 19, PANGO_TAB_LEFT, set_tab_width * 20));
	}

	for (node = open_files; node != NULL; node = g_list_next(node)) {
		struct openfile *file = (struct openfile *)node->data;
		gtk_widget_modify_font(file->textview, font);
	}
}


/*
    Clears the assembler
    output text box
*/

void assembler_output_clear() {

    gtk_text_buffer_set_text(gtk_text_view_get_buffer(output), "", -1);
	gtk_check_menu_item_set_active(outputmenuitem, TRUE);	//make sure it's visible
}


/*
    Appends [text] to the
    assembler output window
*/

void assembler_output_append(gchar *text) {

    GtkTextIter end;
    GtkTextBuffer *textbuffer = gtk_text_view_get_buffer(output);
    gtk_text_buffer_get_end_iter(textbuffer, &end);		//find where the end of the buffer is
    gtk_text_buffer_insert(textbuffer, &end, text, -1);		//insert the text there
	//hack because GTK does some strange stuff with text widgets
	gtk_text_buffer_place_cursor(textbuffer, &end);
	GtkTextMark *mark = gtk_text_buffer_get_insert(textbuffer);
	gtk_text_view_scroll_to_mark(output, mark, 0, FALSE, 0, 0);	//and scroll to the end
}


/*
    Called when cursor is moved -
    updates line and column
    numbers in the status bar
*/

void callback_cursor_moved(GtkTextBuffer *textbuffer, GtkTextIter *iter, GtkTextMark *mark, gpointer data) {

    const gchar *markname = gtk_text_mark_get_name(mark);
    if (markname != NULL && !strcmp(gtk_text_mark_get_name(mark), "insert")) {	//if it's the cursor that moved,

		gchar *statusbar_message = g_strdup_printf("Line %d, Column %d", gtk_text_iter_get_line(iter) + 1, gtk_text_iter_get_line_offset(iter) + 1);
		gtk_statusbar_pop(statusbar, statusbar_context);			//generate a message with the cursor position,
		gtk_statusbar_push(statusbar, statusbar_context, statusbar_message);	//and set that as the statusbar text
		g_free(statusbar_message);
    }
}


/*
    Called when text is changed -
    updates color, handles automatic
    tabs, etc.
*/

/*void callback_text_changed(GtkTextBuffer *widget, gpointer data) {

	printf("Text changed\n");

	GtkTextIter iter;
	GtkTextMark *mark = gtk_text_buffer_get_insert(widget);
	gtk_text_buffer_get_iter_at_mark(widget, &iter, mark);
    callback_cursor_moved(textbuffer, &iter, mark, NULL);	//make sure cursor updates are registered, since the broken "mark-set" signal doesn't do it for us

//    if (!gtk_text_iter_starts_line(&iter))			//get the start and end of the line,
		gtk_text_iter_set_line_offset(&iter, 0);	//TODO: take out this 'if'? (maybe it just wastes time?)

    if (usesyncolor)
		update_line_color(textbuffer, iter);			//and update the syntax coloring for the line
}*/


/*
    Called when text is deleted -
    updates cursor position and syntax highlighting
*/

void callback_text_deleted(GtkTextBuffer *widget, GtkTextIter *start, GtkTextIter *end, gpointer data) {

	GtkTextIter iter;		//updates cursor position (is this really needed?)
	GtkTextMark *mark = gtk_text_buffer_get_insert(widget);
	gtk_text_buffer_get_iter_at_mark(widget, &iter, mark);
	callback_cursor_moved(widget, &iter, mark, NULL);

	if (usesyncolor) {		//runs syntax coloring on line which was changed
		gtk_text_iter_set_line_offset(&iter, 0);	//if parts of two lines were deleted, doesn't bother updating the line where the end of it was
		update_line_color(widget, iter);			// deleted, since that's not going to change how the beginning of the line should look
	}
}


/*
    Called before text is deleted -
    adds action to undo stack
*/

void callback_text_deleted_before(GtkTextBuffer *widget, GtkTextIter *start, GtkTextIter *end, gpointer data) {

    gchar *text = gtk_text_buffer_get_text(widget, start, end, FALSE);
    add_to_undo_stack(1, start, text);	//add action to undo stack
    g_free(text);
}


/*
    Called when text is inserted -
    adds action to undo stack
*/

void callback_text_inserted(GtkTextBuffer *widget, GtkTextIter *start, gchar *text, gint len, gpointer data) {

    add_to_undo_stack(0, start, text);		//add action to undo stack

    if (useautoindent && !strcmp(text, "\n")) {		//if we're supposed to indent automatically and there's a newline,
		GtkTextIter startiter, enditer;			//then get the start of the indent on the last line
		gtk_text_buffer_get_iter_at_line(widget, &startiter, gtk_text_iter_get_line(start) - 1);
		gunichar curr_char;
		enditer = startiter;
		int indentlen = 0;

		while (TRUE) {							//and keep going from there until we find the end of the indent
		    curr_char = gtk_text_iter_get_char(&enditer);
		    if (g_utf8_get_char("\n") == curr_char || g_unichar_isalnum(curr_char) ||
				g_unichar_ispunct(curr_char) ||
				gtk_text_iter_get_offset(start) == gtk_text_iter_get_offset(&enditer))

				break;

			gtk_text_iter_forward_char(&enditer);
		    indentlen++;
		}

		if (gtk_text_iter_get_offset(start) != gtk_text_iter_get_offset(&enditer)) {	//if there actually was an indent on the last line,
			gchar *indenttext = gtk_text_buffer_get_text(widget, &startiter, &enditer, FALSE);	//then get it as text
			GtkTextIter temp;
			gtk_text_buffer_get_iter_at_mark(widget, &temp, gtk_text_buffer_get_insert(widget));
			gtk_text_buffer_insert(widget, &temp, indenttext, -1);				//and insert it at the beginning of this line too
			g_free(indenttext);

			*start = temp;	//quick hack so callback_cursor_moved in the next bit of code
							// will get the right position, hope this doesn't break with
							// different GTK versions - fix by making a copy of the iter
							// that both parts use or something 
		}
    }

	GtkTextIter iter;		//make sure cursor position is updated (doesn't happen by itself?)
	GtkTextMark *mark = gtk_text_buffer_get_insert(widget);
	gtk_text_buffer_get_iter_at_mark(widget, &iter, mark);
	callback_cursor_moved(widget, &iter, mark, NULL);

	if (usesyncolor && strcmp(text, "\n")) {
		int i, len = strlen(text), numlines = 1;	//see how many lines were inserted
		for (i = 0; i < len; i++) {
			if (text[i] == '\n')
				numlines++;
		}

		iter = *start;
		gtk_text_iter_set_line_offset(&iter, 0);
		while (numlines > 0) {						//update the syntax coloring on all of them
			update_line_color(widget, iter);		// this is needed for cases like pasting multiple lines
			gtk_text_iter_backward_line(&iter);		// where they otherwise wouldn't all get colored
			numlines--;
		}
	}
}


/*
    Called when the 'modified' status
    of the text buffers is changed -
    updates title bar and tab labels
*/

void callback_modified_changed(GtkTextBuffer *widget, gpointer data) {

    update_titlebar();
}


/*
    Callback for running
    syntax coloring in background
    after file is opened
*/

gboolean callback_idle_colorall(gpointer data) {

    update_line_color(gtk_text_iter_get_buffer(coloriter[(int)data]), *coloriter[(int)data]);
    if (gtk_text_iter_forward_visible_line(coloriter[(int)data])) {
		return TRUE;
    } else {
		gtk_text_iter_free(coloriter[(int)data]);
		coloriter[(int)data] = NULL;
		return FALSE;
    }
}


/*
    Updates color of a
    single line in [textbuffer],
    starting at [iter]
*/
    
void update_line_color(GtkTextBuffer *textbuffer, GtkTextIter iter) {

    GtkTextIter startiter, enditer;
    startiter = iter;

    if (!gtk_text_iter_ends_line(&iter))				//find the end of the line,
		gtk_text_iter_forward_to_line_end(&iter);
    else
		return;								//(if the start's the same as the end, there's nothing to color)

    enditer = iter;
    
    gtk_text_buffer_remove_all_tags(textbuffer, &startiter, &enditer);	//and remove all existing tags
    gunichar firstchar = gtk_text_iter_get_char(&startiter);
    if (g_unichar_isalnum(firstchar) || firstchar == uchar_uscore) {	//if there isn't whitespace,

		iter = startiter;
		gtk_text_iter_forward_char(&iter);				//it's a label, so set the line to blue
		gunichar uchar;
		while (!g_unichar_isspace((uchar = gtk_text_iter_get_char(&iter))) && uchar != 0 && uchar != uchar_colon) gtk_text_iter_forward_char(&iter);
		if (uchar == uchar_colon)
	    	gtk_text_iter_forward_char(&iter);
		gtk_text_buffer_apply_tag(textbuffer, tag_label, &startiter, &iter);
		startiter = iter;
    }
    
	while (!gtk_text_iter_equal(&startiter, &enditer) && g_unichar_isspace(gtk_text_iter_get_char(&startiter))) gtk_text_iter_forward_char(&startiter);	//skip the whitespace,
	firstchar = gtk_text_iter_get_char(&startiter);
    
	if (firstchar == uchar_dot || firstchar == uchar_hash) {	//and color any directives
    
		iter = startiter;
		gunichar uchar;
		while (!g_unichar_isspace((uchar = gtk_text_iter_get_char(&iter))) && uchar != 0) gtk_text_iter_forward_char(&iter);
		gtk_text_iter_forward_char(&iter);
		gtk_text_buffer_apply_tag(textbuffer, tag_directive, &startiter, &iter);
		startiter = iter;	
	}

	if (gtk_text_iter_forward_search(&startiter, ";", 0, &startiter, NULL, &enditer))	//if there's a ';' in the line,
		gtk_text_buffer_apply_tag(textbuffer, tag_comment, &startiter, &enditer);	//it's a comment, so set it to green
}
