/*
    Z80 Assembly IDE - An IDE for the Z80 line of TI calculators
    Copyright (C) 2006 Donald Straney

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <gtk/gtk.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <unistd.h>
#include <string.h>

#include "asmide.h"
#include "file.h"

gchar *change_filename_extension(gchar*, gchar*);

#define ASM_SUCCESS_STR "=== Assembly succeeded! ==="
#define ASM_FAILURE_STR "=== Assembly failed ==="
#define CONV_FAILURE_STR "=== Program export failed ==="
#define STAT_FAILURE_STR "=== Program stat failed ==="
#define CC_FAILURE_STR "=== Code counter failed ==="
#define SYM_FAILURE_STR "=== Symbol table export failed ==="


/*
    Builds current file
*/

void build_file(struct status *curr_status) {

    struct openfile *file = *(curr_status->curr_file);
    gchar *extension = *(curr_status->output_type);

    char *incdir1 = *(curr_status->incdir1);
    char *incdir2 = *(curr_status->incdir2);
    char *incdir3 = *(curr_status->incdir3);

    if (strlen(file->filename) == 0 || gtk_text_buffer_get_modified(file->textbuffer) == TRUE) {

		GtkWidget *dialog = gtk_message_dialog_new(NULL, 0, GTK_MESSAGE_WARNING, GTK_BUTTONS_YES_NO, "You must save this file\nbefore building it.");
		gtk_message_dialog_format_secondary_text(dialog, "Would you like to save it now?");
		gint result = gtk_dialog_run(dialog);
		gtk_widget_destroy(dialog);

		if (result == GTK_RESPONSE_NO)
		    return;

		if (!save_file(file, curr_status))
		    return;

		gtk_text_buffer_set_modified(file->textbuffer, FALSE);
    }

    gchar *dir = g_path_get_dirname(file->filename);

    assembler_output_clear();
//    assembler_output_append("Starting assembler...\n");

    GError *error = NULL;
    gchar *outfilename = change_filename_extension(file->filename, "bin");

    gchar *argv[] = {"spasm", file->filename, outfilename, NULL, NULL, NULL, NULL, NULL, NULL, NULL};
    int i = 3;
    if (incdir1 != NULL) {
		argv[i++] = "-I";
		argv[i++] = incdir1;
    }
    if (incdir2 != NULL) {
		argv[i++] = "-I";
		argv[i++] = incdir2;
    }
    if (incdir3 != NULL) {
		argv[i++] = "-I";
		argv[i++] = incdir3;
    }
    
    gint exitstatus;
    gchar *output;
    if (!g_spawn_sync(dir, argv, NULL, G_SPAWN_SEARCH_PATH, NULL, NULL, &output, NULL, &exitstatus, &error)) {

		GtkWidget *dialog = gtk_message_dialog_new(NULL, 0, GTK_MESSAGE_ERROR, GTK_BUTTONS_OK, "Couldn't start the assembler.");
		gtk_message_dialog_format_secondary_text(dialog, "%s", error->message);
		gtk_dialog_run(dialog);
		gtk_widget_destroy(dialog);
		g_error_free(error);
		g_free(dir);
		g_free(outfilename);
		assembler_output_append(ASM_FAILURE_STR);
		return;
    }
    
    assembler_output_append(output);
    g_free(output);
    
    if (!WIFEXITED(exitstatus)) {
		GtkWidget *dialog = gtk_message_dialog_new(NULL, 0, GTK_MESSAGE_ERROR, GTK_BUTTONS_OK, "The assembler crashed");
		gtk_message_dialog_format_secondary_text(dialog, "Please report this bug to\nburntfuse@gmail.com along with the code\nthat caused the crash.");
		gtk_dialog_run(dialog);
		gtk_widget_destroy(dialog);
		g_free(outfilename);
		g_free(dir);
		return;
	}

	if (WEXITSTATUS(exitstatus) != 0) {
		assembler_output_append(ASM_FAILURE_STR);
		g_free(outfilename);
		g_free(dir);
		return;
	}

    if (strcasecmp(extension, "bin") != 0) {

//		assembler_output_append("Starting wabbitsign...\n");

		gchar *outfilename2 = change_filename_extension(file->filename, extension);
		gchar *filename = g_path_get_basename(outfilename2);
		gchar *filenamein = g_path_get_basename(outfilename);
		g_free(outfilename2);

		gchar *argv2[] = {"wabbit", filenamein, filename, NULL};	//TODO: quick hack executes wabbitsign in directory of source file - it should really accept full paths
		if (!g_spawn_sync(dir, argv2, NULL, G_SPAWN_SEARCH_PATH, NULL, NULL, &output, NULL, &exitstatus, &error)) {

		    GtkWidget *dialog = gtk_message_dialog_new(NULL, 0, GTK_MESSAGE_ERROR, GTK_BUTTONS_OK, "Couldn't start wabbitsign.");
		    gtk_message_dialog_format_secondary_text(dialog, "%s", error->message);
		    gtk_dialog_run(dialog);
		    gtk_widget_destroy(dialog);
		    g_error_free(error);
		    g_free(filenamein);
		    g_free(filename);
		    g_free(outfilename);
		    g_free(dir);
		    assembler_output_append(CONV_FAILURE_STR);
		    return;
		}

		assembler_output_append(output);
		g_free(output);
		g_free(filenamein);
		g_free(filename);
		remove(outfilename);

	    if (!WIFEXITED(exitstatus)) {
			GtkWidget *dialog = gtk_message_dialog_new(NULL, 0, GTK_MESSAGE_ERROR, GTK_BUTTONS_OK, "Wabbitsign crashed");
			gtk_message_dialog_format_secondary_text(dialog, "Please report this bug to\nburntfuse@gmail.com along with the code\nthat caused the crash.");
			gtk_dialog_run(dialog);
			gtk_widget_destroy(dialog);
			g_free(outfilename);
			g_free(dir);
			return;
		}

		if (WEXITSTATUS(exitstatus) != 0) {
		    assembler_output_append(CONV_FAILURE_STR);
		    g_free(outfilename);
		    g_free(dir);
		    return;
		}
    }

    g_free(dir);
    g_free(outfilename);
    assembler_output_append(ASM_SUCCESS_STR);
}


/*
    Creates symbol table for current file
*/

void create_symbol_table(struct status *curr_status) {

    struct openfile *file = *(curr_status->curr_file);
    char *incdir1 = *(curr_status->incdir1);
    char *incdir2 = *(curr_status->incdir2);
    char *incdir3 = *(curr_status->incdir3);

    if (strlen(file->filename) == 0 || gtk_text_buffer_get_modified(file->textbuffer) == TRUE) {

		GtkWidget *dialog = gtk_message_dialog_new(NULL, 0, GTK_MESSAGE_WARNING, GTK_BUTTONS_YES_NO, "You must save this file\nbefore creating a symbol table.");
		gtk_message_dialog_format_secondary_text(dialog, "Would you like to save it now?");
		gint result = gtk_dialog_run(dialog);
		gtk_widget_destroy(dialog);

		if (result == GTK_RESPONSE_NO)
		    return;

		if (!save_file(file, curr_status))
		    return;

		gtk_text_buffer_set_modified(file->textbuffer, FALSE);
    }

    gchar *dir = g_path_get_dirname(file->filename);
    gchar *outfilename = change_filename_extension(file->filename, "inc");

    FILE *temp = fopen(outfilename, "r");
    if (temp != NULL) {
		fclose(temp);
	
		GtkWidget *dialog = gtk_message_dialog_new(NULL, 0, GTK_MESSAGE_QUESTION, GTK_BUTTONS_YES_NO, "%s already exists", outfilename);
		gtk_message_dialog_format_secondary_text(dialog, "Do you want to replace it?");
		gint result = gtk_dialog_run(dialog);
		gtk_widget_destroy(dialog);
	
		if (result == GTK_RESPONSE_NO) {
		    g_free(dir);
		    g_free(outfilename);
		    return;
		}
    }

    assembler_output_clear();
//    assembler_output_append("Starting assembler...\n");

    GError *error = NULL;

    gchar *argv[] = {"spasm", file->filename, "-L", NULL, NULL, NULL, NULL, NULL, NULL, NULL};
    int i = 3;
    if (incdir1 != NULL) {
		argv[i++] = "-I";
		argv[i++] = incdir1;
    }
    if (incdir2 != NULL) {
		argv[i++] = "-I";
		argv[i++] = incdir2;
    }
    if (incdir3 != NULL) {
		argv[i++] = "-I";
		argv[i++] = incdir3;
    }

    gint exitstatus;
    gchar *output;
    if (!g_spawn_sync(dir, argv, NULL, G_SPAWN_SEARCH_PATH, NULL, NULL, &output, NULL, &exitstatus, &error)) {

		GtkWidget *dialog = gtk_message_dialog_new(NULL, 0, GTK_MESSAGE_ERROR, GTK_BUTTONS_OK, "Couldn't start the assembler.");
		gtk_message_dialog_format_secondary_text(dialog, "%s", error->message);
		gtk_dialog_run(dialog);
		gtk_widget_destroy(dialog);
		g_error_free(error);
		g_free(dir);
		g_free(outfilename);
		assembler_output_append(SYM_FAILURE_STR);
		return;
    }
    
    assembler_output_append(output);
    g_free(output);
    
    g_free(dir);

    if (!WIFEXITED(exitstatus)) {
		GtkWidget *dialog = gtk_message_dialog_new(NULL, 0, GTK_MESSAGE_ERROR, GTK_BUTTONS_OK, "The assembler crashed");
		gtk_message_dialog_format_secondary_text(dialog, "Please report this bug to\nburntfuse@gmail.com along with the code\nthat caused the crash.");
		gtk_dialog_run(dialog);
		gtk_widget_destroy(dialog);
		g_free(outfilename);
		return;
	}

    if (WEXITSTATUS(exitstatus) != 0) {
		assembler_output_append(SYM_FAILURE_STR);
		g_free(outfilename);
		return;
    }

    gchar *filecontents;
    guint filelength;

    if (!open_file_backend(outfilename, &filecontents, &filelength)) {
		g_free(outfilename);
    	return;
    }

    add_file(outfilename, filecontents, filelength);
    g_free(outfilename);
    g_free(filecontents);
}


/*
    Creates code listing
    for current file
*/

void create_listing(struct status *curr_status) {

    struct openfile *file = *(curr_status->curr_file);
    char *incdir1 = *(curr_status->incdir1);
    char *incdir2 = *(curr_status->incdir2);
    char *incdir3 = *(curr_status->incdir3);

    if (strlen(file->filename) == 0 || gtk_text_buffer_get_modified(file->textbuffer) == TRUE) {

		GtkWidget *dialog = gtk_message_dialog_new(NULL, 0, GTK_MESSAGE_WARNING, GTK_BUTTONS_YES_NO, "You must save this file\nbefore generating a code listing.");
		gtk_message_dialog_format_secondary_text(dialog, "Would you like to save it now?");
		gint result = gtk_dialog_run(dialog);
		gtk_widget_destroy(dialog);

		if (result == GTK_RESPONSE_NO)
		    return;

		if (!save_file(file, curr_status))
		    return;

		gtk_text_buffer_set_modified(file->textbuffer, FALSE);
    }

    gchar *dir = g_path_get_dirname(file->filename);
    gchar *outfilename = change_filename_extension(file->filename, "lst");

    FILE *temp = fopen(outfilename, "r");
    if (temp != NULL) {
		fclose(temp);
		GtkWidget *dialog = gtk_message_dialog_new(NULL, 0, GTK_MESSAGE_QUESTION, GTK_BUTTONS_YES_NO, "%s already exists", outfilename);
		gtk_message_dialog_format_secondary_text(dialog, "Do you want to replace it?");
		gint result = gtk_dialog_run(dialog);
		gtk_widget_destroy(dialog);
	
		if (result == GTK_RESPONSE_NO) {
		    g_free(dir);
		    g_free(outfilename);
		    return;
		}
    }

    assembler_output_clear();
//    assembler_output_append("Starting assembler...\n");

    GError *error = NULL;

    gchar *argv[] = {"spasm", file->filename, "-T", NULL, NULL, NULL, NULL, NULL, NULL, NULL};
    int i = 3;
    if (incdir1 != NULL) {
		argv[i++] = "-I";
		argv[i++] = incdir1;
    }
    if (incdir2 != NULL) {
		argv[i++] = "-I";
		argv[i++] = incdir2;
    }
    if (incdir3 != NULL) {
		argv[i++] = "-I";
		argv[i++] = incdir3;
    }

    gint exitstatus;
    gchar *output;
    if (!g_spawn_sync(dir, argv, NULL, G_SPAWN_SEARCH_PATH, NULL, NULL, &output, NULL, &exitstatus, &error)) {

		GtkWidget *dialog = gtk_message_dialog_new(NULL, 0, GTK_MESSAGE_ERROR, GTK_BUTTONS_OK, "Couldn't start the assembler.");
		gtk_message_dialog_format_secondary_text(dialog, "%s", error->message);
		gtk_dialog_run(dialog);
		gtk_widget_destroy(dialog);
		g_error_free(error);
		g_free(dir);
		g_free(outfilename);
		assembler_output_append(ASM_FAILURE_STR);
		return;
    }
    
    assembler_output_append(output);
    g_free(output);
    
    g_free(dir);

    if (!WIFEXITED(exitstatus)) {
		GtkWidget *dialog = gtk_message_dialog_new(NULL, 0, GTK_MESSAGE_ERROR, GTK_BUTTONS_OK, "The assembler crashed");
		gtk_message_dialog_format_secondary_text(dialog, "Please report this bug to\nburntfuse@gmail.com along with the code\nthat caused the crash.");
		gtk_dialog_run(dialog);
		gtk_widget_destroy(dialog);
		g_free(outfilename);
		return;
	}
    
    if (WEXITSTATUS(exitstatus) != 0) {
		assembler_output_append(ASM_FAILURE_STR);
		g_free(outfilename);
		return;
    }

    assembler_output_append(ASM_SUCCESS_STR);
    
    gchar *filecontents;
    guint filelength;

    if (!open_file_backend(outfilename, &filecontents, &filelength)) {
		GtkWidget *dialog = gtk_message_dialog_new(NULL, 0, GTK_MESSAGE_ERROR, GTK_BUTTONS_OK, "Can't open the listing");
		gtk_message_dialog_format_secondary_text(dialog, "This is NOT supposed to happen -\nreport this as a bug to burntfuse@gmail.com");
		gtk_dialog_run(dialog);
		gtk_widget_destroy(dialog);
		g_free(outfilename);
    	return;
    }

    add_file(outfilename, filecontents, filelength);
    g_free(outfilename);
    g_free(filecontents);
}


/*
	Counts size and execution
	time of [code]
*/

void count_code(gchar *code) {

	/* generating random filenames:
		yeah, yeah, I know this isn't 100% bulletproof - here's when it will fail:
		first, for time() to return the same value the button would have to be pressed
		at the exact same time in the two different instances of asmide.
		obviously, the PID couldn't be the same on the same machine for different processes,
		so the /tmp directory (or whatever's in TMPDIR) would have to be shared across
		a network, maybe by NFS.
		so to sum up - if you happen to share your temp directory with another computer by
		network for some brain-damaged reason AND the other system's PID number for asmide
		ends up at the same value AND you press the code counter button on both machines
		within a second of each other, then you should be careful ;-) */
	time_t now;
	time(&now);
	gchar *temp = g_strdup_printf("%d%d", (int)now, getpid());
	gchar *temppath = g_strconcat(g_get_tmp_dir(), G_DIR_SEPARATOR_S, "z80asmide", temp, NULL);
	g_free(temp);

	if (!g_file_set_contents(temppath, code, -1, NULL)) {
		GtkWidget *dialog = gtk_message_dialog_new(NULL, 0, GTK_MESSAGE_ERROR, GTK_BUTTONS_OK, "Can't save file to %s", temppath);
		gtk_message_dialog_format_secondary_text(dialog, "Either you've found a bug or your\ntemp dir is really screwed up");
		gtk_dialog_run(dialog);
		gtk_widget_destroy(dialog);
		g_free(temppath);
		return;
	}

	gint exitstatus;
    gchar *output, *output_err;
	GError *error = NULL;
    gchar *argv[] = {"spasm", temppath, "-C", NULL};
    if (!g_spawn_sync(NULL, argv, NULL, G_SPAWN_SEARCH_PATH, NULL, NULL, &output, &output_err, &exitstatus, &error)) {

		GtkWidget *dialog = gtk_message_dialog_new(NULL, 0, GTK_MESSAGE_ERROR, GTK_BUTTONS_OK, "Couldn't start the assembler.");
		gtk_message_dialog_format_secondary_text(dialog, "%s", error->message);
		gtk_dialog_run(dialog);
		gtk_widget_destroy(dialog);
		g_error_free(error);
		g_unlink(temppath);
		g_free(temppath);
		return;
    }

	g_unlink(temppath);
	g_free(temppath);

    if (!WIFEXITED(exitstatus)) {
		GtkWidget *dialog = gtk_message_dialog_new(NULL, 0, GTK_MESSAGE_ERROR, GTK_BUTTONS_OK, "The assembler crashed");
		gtk_message_dialog_format_secondary_text(dialog, "Please report this bug to\nburntfuse@gmail.com along with the code\nthat caused the crash.");
		gtk_dialog_run(dialog);
		gtk_widget_destroy(dialog);
		g_free(output);
		g_free(output_err);
		return;
	}

    if (WEXITSTATUS(exitstatus) != 0) {
		assembler_output_clear();
		assembler_output_append(output);
		assembler_output_append(CC_FAILURE_STR);
		g_free(output);
		g_free(output_err);
		return;
    }
    
	GtkWidget *dialog = gtk_message_dialog_new(NULL, 0, GTK_MESSAGE_INFO, GTK_BUTTONS_OK, output_err);
	gtk_dialog_run(dialog);
	gtk_widget_destroy(dialog);
    g_free(output);
	g_free(output_err);
}


/*
	Shows stats of code
*/

void print_stats(struct status *curr_status) {

    struct openfile *file = *(curr_status->curr_file);
    char *incdir1 = *(curr_status->incdir1);
    char *incdir2 = *(curr_status->incdir2);
    char *incdir3 = *(curr_status->incdir3);

    if (strlen(file->filename) == 0 || gtk_text_buffer_get_modified(file->textbuffer) == TRUE) {

		GtkWidget *dialog = gtk_message_dialog_new(NULL, 0, GTK_MESSAGE_WARNING, GTK_BUTTONS_YES_NO, "You must save this file\nbefore generating statistics.");
		gtk_message_dialog_format_secondary_text(dialog, "Would you like to save it now?");
		gint result = gtk_dialog_run(dialog);
		gtk_widget_destroy(dialog);

		if (result == GTK_RESPONSE_NO)
		    return;

		if (!save_file(file, curr_status))
		    return;

		gtk_text_buffer_set_modified(file->textbuffer, FALSE);
    }

    gchar *dir = g_path_get_dirname(file->filename);

	gint exitstatus;
    gchar *output, *output_err;
	GError *error = NULL;
    gchar *argv[] = {"spasm", file->filename, "-S", NULL, NULL, NULL, NULL, NULL, NULL, NULL};
    int i = 3;
    if (incdir1 != NULL) {
		argv[i++] = "-I";
		argv[i++] = incdir1;
    }
    if (incdir2 != NULL) {
		argv[i++] = "-I";
		argv[i++] = incdir2;
    }
    if (incdir3 != NULL) {
		argv[i++] = "-I";
		argv[i++] = incdir3;
    }

    if (!g_spawn_sync(dir, argv, NULL, G_SPAWN_SEARCH_PATH, NULL, NULL, &output, &output_err, &exitstatus, &error)) {

		GtkWidget *dialog = gtk_message_dialog_new(NULL, 0, GTK_MESSAGE_ERROR, GTK_BUTTONS_OK, "Couldn't start the assembler.");
		gtk_message_dialog_format_secondary_text(dialog, "%s", error->message);
		gtk_dialog_run(dialog);
		gtk_widget_destroy(dialog);
		g_error_free(error);
		g_free(dir);
		return;
    }

	g_free(dir);

    if (!WIFEXITED(exitstatus)) {
		GtkWidget *dialog = gtk_message_dialog_new(NULL, 0, GTK_MESSAGE_ERROR, GTK_BUTTONS_OK, "The assembler crashed");
		gtk_message_dialog_format_secondary_text(dialog, "Please report this bug to\nburntfuse@gmail.com along with the code\nthat caused the crash.");
		gtk_dialog_run(dialog);
		gtk_widget_destroy(dialog);
		g_free(output);
		g_free(output_err);
		return;
	}

    if (WEXITSTATUS(exitstatus) != 0) {
		assembler_output_clear();
		assembler_output_append(output);
		assembler_output_append(STAT_FAILURE_STR);
		g_free(output);
		g_free(output_err);
		return;
    }

	GtkWidget *dialog = gtk_message_dialog_new(NULL, 0, GTK_MESSAGE_INFO, GTK_BUTTONS_OK, output_err);
	gtk_dialog_run(dialog);
	gtk_widget_destroy(dialog);
    g_free(output);
	g_free(output_err);
}


/*
    Changes filename extension
    of [path] to [extension],
    returns malloc'ed string
*/

gchar *change_filename_extension(gchar *path, gchar *extension) {

    gint i = strlen(path) - 1;

    while (path[i] != '/' && path[i] != '.' && i >= 0) i--;
    
    if (path[i] == '.') {
    
		gchar *temp = g_strndup(path, i + 1);
		gchar *retval = g_strconcat(temp, extension, NULL);
		g_free(temp);
		return retval;

    } else {
    
		return g_strconcat(path, ".", extension, NULL);
    }
}
