/*
    Z80 Assembly IDE - An IDE for the Z80 line of TI calculators
    Copyright (C) 2006 Donald Straney

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <stdlib.h>
#include <gtk/gtk.h>
#include <string.h>

#include "asmide.h"
#include "file.h"
#include "build.h"

void callback_action_setassembleropt_browse(GtkWidget*, gpointer);


/*
    Finds [text], starting at [startiter],
    if found returns TRUE with [resultstartiter]
    set to beginning and [resultenditer] set
    to end, [case_sensitive] sets case sensitivity
*/

gboolean forward_search(GtkTextIter *startiter, gchar *text, gboolean case_sensitive, GtkTextIter *resultstartiter, GtkTextIter *resultenditer) {

    gunichar curr_char = gtk_text_iter_get_char(startiter);
    if (!curr_char)
		return FALSE;
    
    while (1) {
		while (g_unichar_tolower(curr_char) != g_unichar_tolower(*text)) {	//TODO: could be made faster with separate while loops for case sensitive/not case sensitive (less false positives)
	    	gtk_text_iter_forward_char(startiter);
	    	curr_char = gtk_text_iter_get_char(startiter);
	    	if (!curr_char)
				return FALSE;
		}
	
		*resultenditer = *startiter;
		gtk_text_iter_forward_chars(resultenditer, strlen(text));
		gchar *tempstr = gtk_text_iter_get_slice(startiter, resultenditer);

		if ((case_sensitive && !strcmp(tempstr, text)) || (!case_sensitive && !strcasecmp(tempstr, text))) {
		    g_free(tempstr);
		    *resultstartiter = *startiter;
		    return TRUE;
		}

		g_free(tempstr);
		gtk_text_iter_forward_chars(startiter, strlen(text));
		curr_char = gtk_text_iter_get_char(startiter);
    }
}


/*
    Finds [string] in [textbuffer],
    with GtkTextView [text],
    highlights and returns TRUE if
    found, returns FALSE otherwise
*/

gboolean find_text(gchar *string, GtkTextBuffer *textbuffer, GtkTextView *text, gboolean case_sensitive) {

    GtkTextIter iter, startiter, enditer;

    gtk_text_buffer_get_iter_at_mark(textbuffer, &iter, gtk_text_buffer_get_selection_bound(textbuffer));

    if (forward_search(&iter, string, case_sensitive, &startiter, &enditer)) {
	    
        gtk_text_buffer_select_range(textbuffer, &startiter, &enditer);
        gtk_text_view_scroll_to_iter(text, &startiter, 0, FALSE, 0, 0);
		return TRUE;
	
    } else {

        if (gtk_text_iter_is_start(&iter)) {
		    return FALSE;
		} else {
			gtk_text_buffer_get_start_iter(textbuffer, &iter);
			if (forward_search(&iter, string, case_sensitive, &startiter, &enditer)) {
	    		gtk_text_buffer_select_range(textbuffer, &startiter, &enditer);
	    		gtk_text_view_scroll_to_iter(text, &startiter, 0, FALSE, 0, 0);
				return TRUE;
	    	} else {
	    		return FALSE;
	    	}
		}
    }
}


/*
    Creates new file
*/

void callback_action_new(GtkWidget *widget, gpointer data) {

    add_file("", "", 0);			//add a new empty file to the list of open files
}


/*
    Opens file
*/

void callback_action_open(GtkWidget *widget, gpointer data) {

    struct openfile *curr_file = *(((struct status *)data)->curr_file);

    GtkWidget *dialog = gtk_file_chooser_dialog_new("Open File", NULL, GTK_FILE_CHOOSER_ACTION_OPEN, GTK_STOCK_CANCEL, GTK_RESPONSE_CANCEL, GTK_STOCK_OPEN, GTK_RESPONSE_ACCEPT, NULL);

	if (curr_file != NULL && strlen(curr_file->filename) > 0) {
		gchar *path = g_path_get_dirname(curr_file->filename);
		gtk_file_chooser_set_current_folder(dialog, path);
		g_free(path);
	}

    gint response = gtk_dialog_run(dialog);		//create a file dialog and show it

    if (response == GTK_RESPONSE_CANCEL) {
		gtk_widget_destroy(dialog);
		return;						//if the user canceled, don't do anything
    }
    
    gchar *filename = gtk_file_chooser_get_filename(dialog);	//otherwise, get the filename
    gtk_widget_destroy(dialog);

    gchar *filecontents;
    guint filelength;

    if (!open_file_backend(filename, &filecontents, &filelength))	//open it
		return;

    add_file(filename, filecontents, filelength);	//add it to the list of open files

    g_free(filecontents);
    g_free(filename);
}


/*
    Saves file
*/

void callback_action_save(GtkWidget *widget, gpointer data) {

    struct openfile *curr_file = *(((struct status *)data)->curr_file);

    if (!save_file(curr_file, (struct status *)data))			//try to save the file
		return;

    gtk_text_buffer_set_modified(curr_file->textbuffer, FALSE);	//if it worked, show that there aren't any changes to this file,
//    update_titlebar();		//and update the titlebar text
}


/*
    Saves file under
    a different name
*/

void callback_action_saveas(GtkWidget *widget, gpointer data) {		//TODO: combine this with callback_action_save

    struct openfile *curr_file = *(((struct status *)data)->curr_file);

    gchar save_filename[MAX_FILENAME_LEN];

    GtkWidget *dialog = gtk_file_chooser_dialog_new("Save File", NULL, GTK_FILE_CHOOSER_ACTION_SAVE, GTK_STOCK_CANCEL, GTK_RESPONSE_CANCEL, GTK_STOCK_SAVE, GTK_RESPONSE_ACCEPT, NULL);
    gtk_file_chooser_set_do_overwrite_confirmation(dialog, TRUE);

	if (strlen(curr_file->filename) > 0) {
		gchar *path = g_path_get_dirname(curr_file->filename);
		gtk_file_chooser_set_current_folder(dialog, path);
		g_free(path);
	}

    gint response = gtk_dialog_run(dialog);		//show the file dialog

    if (response == GTK_RESPONSE_CANCEL) {
        gtk_widget_destroy(dialog);
        return;
    }
    
    gchar *filename = gtk_file_chooser_get_filename(dialog);	//TODO: check length!
	gtk_widget_destroy(dialog);
    strcpy(save_filename, filename);		//get the filename
    g_free(filename);
/*
    FILE *temp = fopen(save_filename, "r");

    if (temp != NULL) {
	fclose(temp);
	
	GtkWidget *dialog = gtk_message_dialog_new(NULL, 0, GTK_MESSAGE_QUESTION, GTK_BUTTONS_YES_NO, "This file already exists");
	gtk_message_dialog_format_secondary_text(dialog, "Do you want to replace it?");
	gint result = gtk_dialog_run(dialog);
	gtk_widget_destroy(dialog);
	
	if (result == GTK_RESPONSE_NO)
	    return FALSE;
    }
*/
    if (!save_file_backend(curr_file, save_filename))	//and try to save
		return;

    gtk_text_buffer_set_modified(curr_file->textbuffer, FALSE);	//if it worked, show that there aren't any changes to this file,
//	printf("filename is %s\n", save_filename);
    strcpy(curr_file->filename, save_filename);	//save the file name (for if it wasn't defined at first),
    update_titlebar();		//and update the titlebar text
//	printf("filename (real) is %s\n", curr_file->filename);
}


/*
    Reloads open file
*/

void callback_action_revert(GtkWidget *widget, gpointer data) {

    struct openfile *curr_file = *(((struct status *)data)->curr_file);

    if (strlen(curr_file->filename) == 0)	//if the current file doesn't have a name yet, can't revert to saved
		return;

    GtkWidget *dialog = gtk_message_dialog_new(NULL, 0, GTK_MESSAGE_QUESTION, GTK_BUTTONS_YES_NO, "Are you sure you want to\nrevert to the saved version of\n%s?", curr_file->filename);
    gtk_message_dialog_format_secondary_text(dialog, "You will lose any unsaved changes.");
    gint result = gtk_dialog_run(dialog);		//show a dialog for confirmation
    gtk_widget_destroy(dialog);

    if (result == GTK_RESPONSE_NO)		//if the user pressed no, then stop
		return;

    gchar *filecontents;			//otherwise...
    guint filelength;

    if (!open_file_backend(curr_file->filename, &filecontents, &filelength))	//try to open
		return;

    gtk_text_buffer_set_text(curr_file->textbuffer, filecontents, filelength);	//if it worked, set the contents of the text buffer from the saved file
    gtk_text_buffer_set_modified(curr_file->textbuffer, FALSE);

    g_free(filecontents);
    init_file();
//    update_titlebar();
}


/*
    Closes file
*/

void callback_action_close(GtkWidget *widget, gpointer data) {

    remove_file();		//remove the file from the list of open files
}


/*
    Closes all files
*/

void callback_action_closeall(GtkWidget *widget, gpointer data) {

    struct openfile *curr_file = *(((struct status *)data)->curr_file);

    if (curr_file == NULL)
		return;

    remove_file_all();			//close all files
}


/*
    Copies selected text
    to clipboard
*/

void callback_action_copy(GtkWidget *widget, gpointer data) {

    struct openfile *curr_file = *(((struct status *)data)->curr_file);

    GtkTextBuffer *textbuffer = curr_file->textbuffer;	//get a reference to the text buffer
    gtk_text_buffer_copy_clipboard(textbuffer, gtk_clipboard_get(GDK_SELECTION_CLIPBOARD));	//and copy
}


/*
    Cuts selected text
    from clipboard
*/

void callback_action_cut(GtkWidget *widget, gpointer data) {

    struct openfile *curr_file = *(((struct status *)data)->curr_file);

    GtkTextBuffer *textbuffer = curr_file->textbuffer;		//only a n00b moron couldn't understand this after seeing the last routine
    gtk_text_buffer_cut_clipboard(textbuffer, gtk_clipboard_get(GDK_SELECTION_CLIPBOARD), TRUE);
}


/*
    Pastes contents
    of clipboard at cursor
*/

void callback_action_paste(GtkWidget *widget, gpointer data) {

	struct openfile *curr_file = *(((struct status *)data)->curr_file);
	GtkWidget *text = curr_file->textview;

	GtkTextBuffer *textbuffer = curr_file->textbuffer;
	gtk_text_buffer_paste_clipboard(textbuffer, gtk_clipboard_get(GDK_SELECTION_CLIPBOARD), NULL, TRUE);
	gtk_widget_grab_focus(text);
}


/*
    Selects entire contents
    of current file
*/

void callback_action_selectall(GtkWidget *widget, gpointer data) {

	struct openfile *curr_file = *(((struct status *)data)->curr_file);
	GtkWidget *text = curr_file->textview;

	GtkTextBuffer *textbuffer = curr_file->textbuffer;	//first get a reference to the text buffer
	GtkTextIter start, end;
	gtk_text_buffer_get_bounds(textbuffer, &start, &end);	//then get the whole range of text
	gtk_text_buffer_select_range(textbuffer, &start, &end);	//and select it
	gtk_widget_grab_focus(text);
}


/*
    Shows dialog and goes
    to line
*/

void callback_action_gotoline(GtkWidget *widget, gpointer data) {

    struct openfile *curr_file = *(((struct status *)data)->curr_file);
    GtkWidget *text = curr_file->textview;

    GtkWidget *dialog = gtk_dialog_new_with_buttons("Go to line", NULL, GTK_DIALOG_NO_SEPARATOR | GTK_DIALOG_MODAL, GTK_STOCK_OK, GTK_RESPONSE_ACCEPT, GTK_STOCK_CANCEL, GTK_RESPONSE_REJECT, NULL);
    gtk_dialog_set_default_response(dialog, GTK_RESPONSE_ACCEPT);
    gtk_window_set_resizable(dialog, FALSE);
    gtk_box_set_homogeneous(GTK_DIALOG(dialog)->vbox, FALSE);				//create a dialog for getting the line number
    gtk_box_pack_start(GTK_DIALOG(dialog)->vbox, gtk_label_new("Enter a line number"), FALSE, FALSE, 5);
    GtkWidget *linenumentry = gtk_entry_new();	//TODO: set max length
    gtk_box_pack_start(GTK_DIALOG(dialog)->vbox, linenumentry, TRUE, TRUE, 5);
    gtk_widget_show_all(dialog);
    gint result = gtk_dialog_run(dialog);

    if (result == GTK_RESPONSE_REJECT) {
		gtk_widget_destroy(dialog);
		return;
    }

    gint linenum = atoi(gtk_entry_get_text(linenumentry));	//convert the line number to an int from a string

    gtk_widget_destroy(dialog);

    GtkTextIter iter;				//and set the cursor to that line
    gtk_text_buffer_get_iter_at_line(curr_file->textbuffer, &iter, linenum - 1);
    gtk_text_buffer_place_cursor(curr_file->textbuffer, &iter);
    gtk_text_view_scroll_to_iter(text, &iter, 0, TRUE, 0, 0);	//make the text box scroll to it
    gtk_widget_grab_focus(text);
}


/*
    Shows dialog and goes
    to label
*/

void callback_action_gotolabel(GtkWidget *widget, gpointer data) {	//TODO: add case-sensitive option

    struct openfile *curr_file = *(((struct status *)data)->curr_file);
    GtkWidget *text = curr_file->textview;

    GtkWidget *dialog = gtk_dialog_new_with_buttons("Go to label", NULL, GTK_DIALOG_NO_SEPARATOR | GTK_DIALOG_MODAL, GTK_STOCK_OK, GTK_RESPONSE_ACCEPT, GTK_STOCK_CANCEL, GTK_RESPONSE_REJECT, NULL);
    gtk_dialog_set_default_response(dialog, GTK_RESPONSE_ACCEPT);
    gtk_window_set_resizable(dialog, FALSE);
    gtk_box_set_homogeneous(GTK_DIALOG(dialog)->vbox, FALSE);				//create a dialog for getting the line number
    gtk_box_pack_start(GTK_DIALOG(dialog)->vbox, gtk_label_new("Enter a label name"), FALSE, FALSE, 5);
    GtkWidget *linenumentry = gtk_entry_new();	//TODO: set max length
    gtk_box_pack_start(GTK_DIALOG(dialog)->vbox, linenumentry, TRUE, TRUE, 5);
    gtk_widget_show_all(dialog);
    gint result = gtk_dialog_run(dialog);

    if (result == GTK_RESPONSE_REJECT) {
		gtk_widget_destroy(dialog);
		return;
    }

    gboolean found_label = FALSE;
    GtkTextIter iter;
    gtk_text_buffer_get_start_iter(curr_file->textbuffer, &iter);		//start at the beginning of the buffer...
    gchar *search_text = gtk_entry_get_text(linenumentry);			//get the text to search for

    while (TRUE) {
    
		gunichar firstchar = gtk_text_iter_get_char(&iter);

		if (g_unichar_isalnum(firstchar) || firstchar == g_utf8_get_char("_")) {	//if there's a label,

		    GtkTextIter startiter, enditer;
		    startiter = iter;
		    gtk_text_iter_forward_char(&iter);
		    while (!g_unichar_isspace(gtk_text_iter_get_char(&iter)) && gtk_text_iter_get_char(&iter) != 0 && gtk_text_iter_get_char(&iter) != g_utf8_get_char(":")) gtk_text_iter_forward_char(&iter);	//TODO: optimize (only access iterator char once)
		    enditer = iter;
		    iter = startiter;

		    gchar *label_text = gtk_text_buffer_get_text(curr_file->textbuffer, &startiter, &enditer, FALSE);	//now get the label text
		    gint result = strcmp(label_text, search_text);	//are they the same?
		    g_free(label_text);
	    
		    if (result == 0) {					//if they are, then exit the loop
				found_label = TRUE;
				break;
		    }
		}

		gboolean result;					//now try to move to the next line
		if (gtk_text_iter_ends_line(&iter))
		    result = gtk_text_iter_forward_char(&iter);
		else
		    result = gtk_text_iter_forward_line(&iter);
	    
		if (!result)						//if we're at the end of the buffer, then the label hasn't been found...
		    break;
    }

    if (found_label) {							//if the label was found,

		gtk_text_buffer_place_cursor(curr_file->textbuffer, &iter);			//set the cursor to that line
		gtk_text_view_scroll_to_iter(text, &iter, 0, TRUE, 0, 0);	//make the text box scroll to it
		gtk_widget_grab_focus(text);

    } else {

		GtkWidget *dialog = gtk_message_dialog_new(NULL, 0, GTK_MESSAGE_ERROR, GTK_BUTTONS_OK, "Couldn't find label \'%s\'", search_text);
		gtk_dialog_run(dialog);				//otherwise, show an error
		gtk_widget_destroy(dialog);
    }

    gtk_widget_destroy(dialog);
}


/*
    Shows "about" dialog
*/

void callback_action_about(GtkWidget *widget, gpointer data) {

    GtkWidget *dialog = gtk_about_dialog_new();
    gtk_about_dialog_set_name(dialog, "Z80 Assembly IDE");
    gtk_about_dialog_set_version(dialog, VERSION);
    gtk_about_dialog_set_copyright(dialog, "Copyright Donald Straney 2007");
    gtk_about_dialog_set_comments(dialog, "\nAn IDE for the Z80 line of TI calculators\n\nSend comments/suggestions/flames\nto burntfuse@gmail.com\n\nA product of Revolution Software - \nsee our website at www.revsoft.org");
    gchar *license;
    if (g_file_get_contents(PREFIX"/share/doc/z80asmide/LICENSE", &license, NULL, NULL)) {
		gtk_about_dialog_set_license(dialog, license);
		g_free(license);
    }
    gchar *authors[] = {"Editor: Donald Straney", "SPASM (assembler): Spencer Putt", "Wabbitsign (app signer/export utility): James Montelongo, Spencer Putt", NULL};
    gtk_about_dialog_set_authors(dialog, authors);
    gtk_about_dialog_set_logo_icon_name(dialog, "z80asmide");
    gtk_dialog_run(GTK_DIALOG(dialog));
    gtk_widget_destroy(GTK_WIDGET(dialog));
}


/*
    Builds program
*/

void callback_action_build(GtkWidget *widget, gpointer data) {

    build_file((struct status *)data);
}


/*
    Creates symbol table
    for current program
*/

void callback_action_createsymtable(GtkWidget *widget, gpointer data) {

    create_symbol_table((struct status *)data);
}


/*
    Generates code listing
*/

void callback_action_createlisting(GtkWidget *widget, gpointer data) {

    create_listing((struct status *)data);
}


/*
	Runs code counter
*/

void callback_action_countcode(GtkWidget *widget, gpointer data) {

	struct openfile *curr_file = *(((struct status *)data)->curr_file);
	GtkTextBuffer *textbuffer = curr_file->textbuffer;

	GtkTextIter start, end;
	if (!gtk_text_buffer_get_selection_bounds(textbuffer, &start, &end))
		return;
	gchar *selected_text = gtk_text_buffer_get_text(textbuffer, &start, &end, FALSE);
	count_code(selected_text);
	g_free(selected_text);
}


/*
	Prints program stats
*/

void callback_action_printstats(GtkWidget *widget, gpointer data) {

	print_stats((struct status *)data);
}


/*
    Sets output program type
*/

void callback_action_setoutputtype(GtkWidget *widget, gpointer data) {

    gchar **output_type = ((struct status *)data)->output_type;

    GtkWidget *dialog = gtk_dialog_new_with_buttons("Set output type", NULL, GTK_DIALOG_NO_SEPARATOR | GTK_DIALOG_MODAL, GTK_STOCK_OK, GTK_RESPONSE_ACCEPT, GTK_STOCK_CANCEL, GTK_RESPONSE_REJECT, NULL);
    gtk_dialog_set_default_response(dialog, GTK_RESPONSE_ACCEPT);
    gtk_window_set_resizable(dialog, FALSE);
    gtk_box_set_homogeneous(GTK_DIALOG(dialog)->vbox, FALSE);				//create a dialog for listing the possibilities

    GtkListStore *list = gtk_list_store_new(2, G_TYPE_STRING, G_TYPE_STRING);	//create the treeview widget and list for displaying them
    GtkWidget *treeview = gtk_tree_view_new_with_model(list);
    GtkCellRenderer *textrenderer = gtk_cell_renderer_text_new();
    GtkTreeViewColumn *column = gtk_tree_view_column_new_with_attributes("", textrenderer, "text", 0, NULL);
    gtk_tree_view_append_column(treeview, column);
    column = gtk_tree_view_column_new_with_attributes("", textrenderer, "text", 1, NULL);
    gtk_tree_view_append_column(treeview, column);
    gtk_tree_view_set_headers_visible(treeview, FALSE);
    gtk_tree_view_set_rules_hint(treeview, TRUE);
    gtk_tree_selection_set_mode(gtk_tree_view_get_selection(treeview), GTK_SELECTION_BROWSE);
    
    GtkTreeIter iter, selection;
    gtk_list_store_insert_with_values(list, &iter, 0, 0, "TI-86 program", 1, "86p", -1);	//fill the list
    if (!strcasecmp(*output_type, "86p"))
		selection = iter;
    gtk_list_store_insert_with_values(list, &iter, 0, 0, "TI-86 string", 1, "86s", -1);
    if (!strcasecmp(*output_type, "86s"))
		selection = iter;
    gtk_list_store_insert_with_values(list, &iter, 0, 0, "TI-85 program", 1, "85p", -1);
    if (!strcasecmp(*output_type, "85p"))
		selection = iter;
    gtk_list_store_insert_with_values(list, &iter, 0, 0, "TI-85 string", 1, "85s", -1);
    if (!strcasecmp(*output_type, "85s"))
		selection = iter;
    gtk_list_store_insert_with_values(list, &iter, 0, 0, "TI-83/84/+/SE program", 1, "8xp", -1);
    if (!strcasecmp(*output_type, "8xp"))
		selection = iter;
    gtk_list_store_insert_with_values(list, &iter, 0, 0, "TI-83/84/+/SE application", 1, "8xk", -1);
    if (!strcasecmp(*output_type, "8xk"))
		selection = iter;
    gtk_list_store_insert_with_values(list, &iter, 0, 0, "TI-83 program", 1, "83p", -1);
    if (!strcasecmp(*output_type, "83p"))
		selection = iter;
    gtk_list_store_insert_with_values(list, &iter, 0, 0, "TI-82 program", 1, "82p", -1);
    if (!strcasecmp(*output_type, "82p"))
		selection = iter;
    gtk_list_store_insert_with_values(list, &iter, 0, 0, "TI-73 program", 1, "73p", -1);
    if (!strcasecmp(*output_type, "73p"))
		selection = iter;
    gtk_list_store_insert_with_values(list, &iter, 0, 0, "Binary file", 1, "bin", -1);
    if (!strcasecmp(*output_type, "bin"))
		selection = iter;

    gtk_tree_selection_select_iter(gtk_tree_view_get_selection(treeview), &selection);	//select the one that's being used
    
    GtkWidget *frame = gtk_frame_new(NULL);			//now pack everything into the dialog
    gtk_frame_set_shadow_type(frame, GTK_SHADOW_IN);
    gtk_container_add(frame, treeview);
    gtk_container_add(GTK_DIALOG(dialog)->vbox, frame);

    gtk_widget_show_all(dialog);
    gint response = gtk_dialog_run(dialog);			//run it, and get the user's response

    if (response == GTK_RESPONSE_REJECT) {			//if the user canceled,
        gtk_widget_destroy(dialog);
		return;							//then exit without changing anything
    }

    gtk_tree_selection_get_selected(gtk_tree_view_get_selection(treeview), NULL, &selection);	//otherwise, get the row of the selected type,
    g_free(*output_type);								//free the old output type string,
    gtk_tree_model_get(list, &selection, 1, output_type, -1);			//and get the filename extension from that row
    gtk_widget_destroy(dialog);
}


/*
    Shows editor
    settings dialog
*/

void callback_action_seteditoropt(GtkWidget *widget, gpointer data) {

    gboolean *usesyncolor = ((struct status *)data)->usesyncolor;
    gboolean *usewordwrap = ((struct status *)data)->usewordwrap;
    gboolean *useautoindent = ((struct status *)data)->useautoindent;
    gint *tabwidth = ((struct status *)data)->tabwidth;
    gchar **color_label = ((struct status *)data)->color_label;
    gchar **color_comment = ((struct status *)data)->color_comment;
    gchar **color_directive = ((struct status *)data)->color_directive;
    PangoFontDescription **font = ((struct status *)data)->font;

    GtkWidget *dialog = gtk_dialog_new_with_buttons("Editor options", NULL, GTK_DIALOG_NO_SEPARATOR | GTK_DIALOG_MODAL, GTK_STOCK_OK, GTK_RESPONSE_ACCEPT, GTK_STOCK_CANCEL, GTK_RESPONSE_REJECT, NULL);
    gtk_dialog_set_default_response(dialog, GTK_RESPONSE_ACCEPT);
    gtk_window_set_resizable(dialog, FALSE);
    gtk_box_set_homogeneous(GTK_DIALOG(dialog)->vbox, FALSE);		//create a dialog for the editor options

    GdkColor gdk_color_label, gdk_color_comment, gdk_color_directive;
    gdk_color_parse(*color_label, &gdk_color_label);
    gdk_color_parse(*color_comment, &gdk_color_comment);
    gdk_color_parse(*color_directive, &gdk_color_directive);
    GtkWidget *label_color_button = gtk_color_button_new_with_color(&gdk_color_label);	//create color buttons and init them with the existing colors
    GtkWidget *comment_color_button = gtk_color_button_new_with_color(&gdk_color_comment);
    GtkWidget *directive_color_button = gtk_color_button_new_with_color(&gdk_color_directive);

    GtkWidget *tabwidth_text = gtk_entry_new();
    gtk_entry_set_width_chars(tabwidth_text, 2);
    gtk_entry_set_max_length(tabwidth_text, 1);
    GtkWidget *usesyncolor_cbutton = gtk_check_button_new();
    GtkWidget *usewordwrap_cbutton = gtk_check_button_new();
    GtkWidget *useautoindent_cbutton = gtk_check_button_new();
    
    gchar *fontname = pango_font_description_to_string(*font);
    GtkWidget *font_button = gtk_font_button_new_with_font(fontname);	//create the font button and init it with the current font
    g_free(fontname);

    GtkWidget *table = gtk_table_new(8, 2, FALSE);
    gtk_container_add(GTK_DIALOG(dialog)->vbox, table);
    gtk_table_attach(table, gtk_label_new("Tab width"), 0, 1, 0, 1, 0, 0, 5, 3);
    gtk_table_attach(table, gtk_label_new("Use word wrap"), 0, 1, 1, 2, 0, 0, 5, 3);
    gtk_table_attach(table, gtk_label_new("Editor font"), 0, 1, 2, 3, 0, 0, 5, 3);
    gtk_table_attach(table, gtk_label_new("Use automatic indent"), 0, 1, 3, 4, 0, 0, 5, 3);
    gtk_table_attach(table, gtk_label_new("Use syntax coloring"), 0, 1, 4, 5, 0, 0, 5, 3);
    gtk_table_attach(table, gtk_label_new("Label color"), 0, 1, 5, 6, 0, 0, 5, 3);
    gtk_table_attach(table, gtk_label_new("Comment color"), 0, 1, 6, 7, 0, 0, 5, 3);
    gtk_table_attach(table, gtk_label_new("Directive color"), 0, 1, 7, 8, 0, 0, 5, 3);
    gtk_table_attach(table, tabwidth_text, 1, 2, 0, 1, GTK_EXPAND | GTK_FILL, 0, 5, 3);
    gtk_table_attach(table, usewordwrap_cbutton, 1, 2, 1, 2, GTK_EXPAND | GTK_FILL, 0, 5, 3);
    gtk_table_attach(table, font_button, 1, 2, 2, 3, GTK_EXPAND | GTK_FILL, 0, 5, 3);
    gtk_table_attach(table, useautoindent_cbutton, 1, 2, 3, 4, GTK_EXPAND | GTK_FILL, 0, 5, 3);
    gtk_table_attach(table, usesyncolor_cbutton, 1, 2, 4, 5, GTK_EXPAND | GTK_FILL, 0, 5, 3);
    gtk_table_attach(table, label_color_button, 1, 2, 5, 6, GTK_EXPAND | GTK_FILL, 0, 5, 3);
    gtk_table_attach(table, comment_color_button, 1, 2, 6, 7, GTK_EXPAND | GTK_FILL, 0, 5, 3);
    gtk_table_attach(table, directive_color_button, 1, 2, 7, 8, GTK_EXPAND | GTK_FILL, 0, 5, 3);

    gchar *width_str = g_strdup_printf("%d", *tabwidth);
    gtk_entry_set_text(tabwidth_text, width_str);
    g_free(width_str);
    gtk_toggle_button_set_active(usesyncolor_cbutton, *usesyncolor);
    gtk_toggle_button_set_active(usewordwrap_cbutton, *usewordwrap);
    gtk_toggle_button_set_active(useautoindent_cbutton, *useautoindent);

    gtk_widget_show_all(dialog);
    gint result = gtk_dialog_run(dialog);
    
    if (result == GTK_RESPONSE_REJECT) {
		gtk_widget_destroy(dialog);
		return;
    }

    gtk_color_button_get_color(label_color_button, &gdk_color_label);			//get the colors,
    gtk_color_button_get_color(comment_color_button, &gdk_color_comment);
    gtk_color_button_get_color(directive_color_button, &gdk_color_directive);
    g_free(*color_label);							//convert them to strings,
    g_free(*color_comment);
    g_free(*color_directive);
    *color_label = g_strdup_printf("#%.2X%.2X%.2X", (gdk_color_label.red & 0xff00) >> 8, (gdk_color_label.green & 0xff00) >> 8, (gdk_color_label.blue & 0xff00) >> 8);
    *color_comment = g_strdup_printf("#%.2X%.2X%.2X", (gdk_color_comment.red & 0xff00) >> 8, (gdk_color_comment.green & 0xff00) >> 8, (gdk_color_comment.blue & 0xff00) >> 8);
    *color_directive = g_strdup_printf("#%.2X%.2X%.2X", (gdk_color_directive.red & 0xff00) >> 8, (gdk_color_directive.green & 0xff00) >> 8, (gdk_color_directive.blue & 0xff00) >> 8);

    pango_font_description_free(*font);
    *font = pango_font_description_from_string(gtk_font_button_get_font_name(font_button));	//create a font from font name

    *tabwidth = atoi(gtk_entry_get_text(tabwidth_text));
    refresh_tabs();

    *usewordwrap = gtk_toggle_button_get_active(usewordwrap_cbutton);
    *useautoindent = gtk_toggle_button_get_active(useautoindent_cbutton);
    int oldusesyncolor = *usesyncolor;
    *usesyncolor = gtk_toggle_button_get_active(usesyncolor_cbutton);

    if (oldusesyncolor != *usesyncolor)		//if the "use syntax coloring" setting has changed,
	set_unset_syncolor();			//refresh all open buffers

    gtk_widget_destroy(dialog);
    
    refresh_tags();							//and refresh the text properties
    refresh_wordwrap();
}


/*
    Shows assembler
    settings dialog
*/

void callback_action_setassembleropt(GtkWidget *widget, gpointer data) {

    gchar **incdir1 = ((struct status *)data)->incdir1;
    gchar **incdir2 = ((struct status *)data)->incdir2;
    gchar **incdir3 = ((struct status *)data)->incdir3;

    GtkWidget *dialog = gtk_dialog_new_with_buttons("Assembler options", NULL, GTK_DIALOG_NO_SEPARATOR | GTK_DIALOG_MODAL, GTK_STOCK_OK, GTK_RESPONSE_ACCEPT, GTK_STOCK_CANCEL, GTK_RESPONSE_REJECT, NULL);
    gtk_dialog_set_default_response(dialog, GTK_RESPONSE_ACCEPT);
    gtk_window_set_resizable(dialog, FALSE);
    gtk_box_set_homogeneous(GTK_DIALOG(dialog)->vbox, FALSE);		//create a dialog for the assembler options

    GtkWidget *hbox = gtk_hbox_new(0, FALSE);			//pack everything into the dialog
    gtk_container_add(GTK_DIALOG(dialog)->vbox, hbox);
    GtkWidget *labelvbox = gtk_vbox_new(0, TRUE);
    GtkWidget *controlvbox = gtk_vbox_new(0, TRUE);
    gtk_box_pack_start(hbox, labelvbox, FALSE, FALSE, 7);
    gtk_box_pack_start(hbox, controlvbox, FALSE, FALSE, 7);

//    GtkWidget *cslabels_cbutton = gtk_check_button_new();		//AGGHHH...too many friggin' widgets...
    GtkWidget *chbox1 = gtk_hbox_new(0, FALSE);
    GtkWidget *incdir1_text = gtk_entry_new();
    gtk_entry_set_width_chars(incdir1_text, 30);
    GtkWidget *incdir_button = gtk_button_new_with_label("Browse");
    g_signal_connect(incdir_button, "clicked", callback_action_setassembleropt_browse, incdir1_text);
    gtk_box_pack_start(chbox1, incdir1_text, FALSE, FALSE, 2);
    gtk_box_pack_start(chbox1, incdir_button, FALSE, FALSE, 2);
    GtkWidget *chbox2 = gtk_hbox_new(0, FALSE);
    GtkWidget *incdir2_text = gtk_entry_new();
    gtk_entry_set_width_chars(incdir2_text, 30);
    incdir_button = gtk_button_new_with_label("Browse");
    g_signal_connect(incdir_button, "clicked", callback_action_setassembleropt_browse, incdir2_text);
    gtk_box_pack_start(chbox2, incdir2_text, FALSE, FALSE, 2);
    gtk_box_pack_start(chbox2, incdir_button, FALSE, FALSE, 2);
    GtkWidget *chbox3 = gtk_hbox_new(0, FALSE);
    GtkWidget *incdir3_text = gtk_entry_new();
    gtk_entry_set_width_chars(incdir3_text, 30);
    incdir_button = gtk_button_new_with_label("Browse");
    g_signal_connect(incdir_button, "clicked", callback_action_setassembleropt_browse, incdir3_text);
    gtk_box_pack_start(chbox3, incdir3_text, FALSE, FALSE, 2);
    gtk_box_pack_start(chbox3, incdir_button, FALSE, FALSE, 2);

/*    gtk_box_pack_start(labelvbox, gtk_label_new("Case-sensitive labels"), TRUE, TRUE, 2);
    gtk_box_pack_start(controlvbox, cslabels_cbutton, TRUE, TRUE, 2); */
    gtk_box_pack_start(labelvbox, gtk_label_new("Include directory 1"), TRUE, TRUE, 2);
    gtk_box_pack_start(controlvbox, chbox1, TRUE, TRUE, 2);
    gtk_box_pack_start(labelvbox, gtk_label_new("Include directory 2"), TRUE, TRUE, 2);
    gtk_box_pack_start(controlvbox, chbox2, TRUE, TRUE, 2);
    gtk_box_pack_start(labelvbox, gtk_label_new("Include directory 3"), TRUE, TRUE, 2);
    gtk_box_pack_start(controlvbox, chbox3, TRUE, TRUE, 2);

    if (*incdir1 != NULL)
        gtk_entry_set_text(incdir1_text, *incdir1);
    if (*incdir2 != NULL)
        gtk_entry_set_text(incdir2_text, *incdir2);
    if (*incdir3 != NULL)
        gtk_entry_set_text(incdir3_text, *incdir3);

    gtk_widget_show_all(dialog);
    gint result = gtk_dialog_run(dialog);
    
    if (result == GTK_RESPONSE_REJECT) {
		gtk_widget_destroy(dialog);
		return;
    }

    if (*incdir1 != NULL)
        g_free(*incdir1);
    if (*incdir2 != NULL)
        g_free(*incdir2);
    if (*incdir3 != NULL)
        g_free(*incdir3);

    if (strlen(gtk_entry_get_text(incdir1_text)) > 0)
        *incdir1 = g_strdup(gtk_entry_get_text(incdir1_text));
    else
		*incdir1 = NULL;
    if (strlen(gtk_entry_get_text(incdir2_text)) > 0)
        *incdir2 = g_strdup(gtk_entry_get_text(incdir2_text));
    else
		*incdir2 = NULL;
    if (strlen(gtk_entry_get_text(incdir3_text)) > 0)
        *incdir3 = g_strdup(gtk_entry_get_text(incdir3_text));
    else
		*incdir3 = NULL;

    gtk_widget_destroy(dialog);
}

void callback_action_setassembleropt_browse(GtkWidget *widget, gpointer data) {

    GtkWidget *dialog = gtk_file_chooser_dialog_new("Select a folder", NULL, GTK_FILE_CHOOSER_ACTION_SELECT_FOLDER, "Select", GTK_RESPONSE_ACCEPT, "Cancel", GTK_RESPONSE_REJECT, NULL);
    if (gtk_dialog_run(dialog) == GTK_RESPONSE_REJECT) {
		gtk_widget_destroy(dialog);
		return;
    }
    
    gtk_entry_set_text(GTK_ENTRY(data), gtk_file_chooser_get_filename(dialog));
    gtk_widget_destroy(dialog);
}


/*
    Shows 'find' dialog
*/

void callback_action_find(GtkWidget *widget, gpointer data) {

    struct openfile *curr_file = *(((struct status *)data)->curr_file);
    GtkWidget *text = curr_file->textview;

    GtkWidget *dialog = gtk_dialog_new_with_buttons("Find text", NULL, GTK_DIALOG_NO_SEPARATOR | GTK_DIALOG_MODAL, GTK_STOCK_FIND, GTK_RESPONSE_ACCEPT, GTK_STOCK_CANCEL, GTK_RESPONSE_REJECT, NULL);
    gtk_dialog_set_default_response(dialog, GTK_RESPONSE_ACCEPT);
    gtk_window_set_resizable(dialog, FALSE);
    gtk_box_set_homogeneous(GTK_DIALOG(dialog)->vbox, FALSE);		//create a dialog

    GtkWidget *hbox = gtk_hbox_new(FALSE, 2);
    GtkWidget *entry = gtk_entry_new();				//and as usual, create widgets
    GtkWidget *casebutton = gtk_check_button_new_with_label("Case-sensitive");

    gtk_box_pack_start(hbox, gtk_label_new("Search for:"), FALSE, FALSE, 2);
    gtk_box_pack_start(hbox, entry, FALSE, FALSE, 2);
    gtk_box_pack_start(GTK_DIALOG(dialog)->vbox, hbox, FALSE, FALSE, 2);	//and pack them into the dialog
    gtk_box_pack_start(GTK_DIALOG(dialog)->vbox, casebutton, FALSE, FALSE, 2);
    
    gtk_widget_show_all(dialog);
    gint response = gtk_dialog_run(dialog);

    while (response == GTK_RESPONSE_ACCEPT) {			//show the dialog until the user presses cancel
    
		if (!find_text(gtk_entry_get_text(entry), curr_file->textbuffer, text, gtk_toggle_button_get_active(casebutton))) {

		    GtkWidget *dialog = gtk_message_dialog_new(NULL, GTK_DIALOG_MODAL, GTK_MESSAGE_INFO, GTK_BUTTONS_OK, "Couldn't find \'%s\'", gtk_entry_get_text(entry));
		    gtk_dialog_run(dialog);
		    gtk_widget_destroy(dialog);
		}
		response = gtk_dialog_run(dialog);
    }

    gtk_widget_destroy(dialog);
}


/*
    Shows 'replace' dialog
*/

void callback_action_replace(GtkWidget *widget, gpointer data) {

    //BUG: check for 'find' string in 'replace' string, refuse to 'replace all' if found (otherwise infinite loop)
    struct openfile *curr_file = *(((struct status *)data)->curr_file);
    GtkWidget *text = curr_file->textview;

    GtkWidget *dialog = gtk_dialog_new_with_buttons("Find/replace text", NULL, GTK_DIALOG_NO_SEPARATOR | GTK_DIALOG_MODAL, GTK_STOCK_FIND_AND_REPLACE, GTK_RESPONSE_ACCEPT, GTK_STOCK_CANCEL, GTK_RESPONSE_REJECT, NULL);
    gtk_dialog_set_default_response(dialog, GTK_RESPONSE_ACCEPT);
    gtk_window_set_resizable(dialog, FALSE);
    gtk_box_set_homogeneous(GTK_DIALOG(dialog)->vbox, FALSE);		//create a dialog

    GtkWidget *hbox = gtk_hbox_new(FALSE, 2);
    GtkWidget *vbox = gtk_vbox_new(TRUE, 0);
    GtkWidget *vbox2 = gtk_vbox_new(TRUE, 0);
    GtkWidget *findtext = gtk_entry_new();				//and as usual, create widgets
    GtkWidget *replacetext = gtk_entry_new();
    GtkWidget *casebutton = gtk_check_button_new_with_label("Case-sensitive");
    GtkWidget *allbutton = gtk_check_button_new_with_label("Replace all");

    gtk_box_pack_start(vbox, gtk_label_new("Search for:"), FALSE, FALSE, 2);
    gtk_box_pack_start(vbox2, findtext, FALSE, FALSE, 2);
    gtk_box_pack_start(vbox, gtk_label_new("Replace with:"), FALSE, FALSE, 2);
    gtk_box_pack_start(vbox2, replacetext, FALSE, FALSE, 2);
    gtk_box_pack_start(hbox, vbox, FALSE, FALSE, 2);
    gtk_box_pack_start(hbox, vbox2, FALSE, FALSE, 2);
    gtk_box_pack_start(GTK_DIALOG(dialog)->vbox, hbox, FALSE, FALSE, 2);	//and pack them into the dialog
    gtk_box_pack_start(GTK_DIALOG(dialog)->vbox, casebutton, FALSE, FALSE, 2);
    gtk_box_pack_start(GTK_DIALOG(dialog)->vbox, allbutton, FALSE, FALSE, 2);
    
    gtk_widget_show_all(dialog);
    gint response = gtk_dialog_run(dialog);

    while (response == GTK_RESPONSE_ACCEPT) {				//if the user pressed 'find'...
    
		if (gtk_toggle_button_get_active(allbutton)) {			//if we're supposed to replace all of them,

		    gint num_replaced = 0;						//init 'number of replacements' counter

		    while (find_text(gtk_entry_get_text(findtext), curr_file->textbuffer, text, gtk_toggle_button_get_active(casebutton))) {	//then replace each occurrence
				num_replaced++;
				gtk_text_buffer_delete_selection(curr_file->textbuffer, FALSE, TRUE);
				gtk_text_buffer_insert_at_cursor(curr_file->textbuffer, gtk_entry_get_text(replacetext), -1);
			}

			GtkWidget *dialog = gtk_message_dialog_new(NULL, GTK_DIALOG_MODAL, GTK_MESSAGE_INFO, GTK_BUTTONS_OK, "Done, %d occurrences replaced", num_replaced);
			gtk_dialog_run(dialog);					//and show a summary when done
			gtk_widget_destroy(dialog);

		} else {

	    	if (find_text(gtk_entry_get_text(findtext), curr_file->textbuffer, text, gtk_toggle_button_get_active(casebutton))) {	//otherwise, try to find the next occurrence
	    
				GtkWidget *dialog = gtk_message_dialog_new(NULL, GTK_DIALOG_MODAL, GTK_MESSAGE_QUESTION, GTK_BUTTONS_YES_NO, "Replace \'%s\' with \'%s\'?", gtk_entry_get_text(findtext), gtk_entry_get_text(replacetext));
				gint response = gtk_dialog_run(dialog);
				gtk_widget_destroy(dialog);
		
				if (response == GTK_RESPONSE_YES) {				//if the user wants it replaced, then replace it
		    
				    gtk_text_buffer_delete_selection(curr_file->textbuffer, FALSE, TRUE);
				    gtk_text_buffer_insert_at_cursor(curr_file->textbuffer, gtk_entry_get_text(replacetext), -1);
				}

	    	} else {			//if it wasn't found, show an error

				GtkWidget *dialog = gtk_message_dialog_new(NULL, GTK_DIALOG_MODAL, GTK_MESSAGE_INFO, GTK_BUTTONS_OK, "Couldn't find \'%s\'", gtk_entry_get_text(findtext));
				gtk_dialog_run(dialog);
				gtk_widget_destroy(dialog);
	    	}
		}
		response = gtk_dialog_run(dialog);
    }
    
    gtk_widget_destroy(dialog);
}


/*
    Opens manual in browser
*/

void callback_action_showhelp(GtkWidget *widget, gpointer data) {

    gchar *browsers[] = {"gnome-open", "firefox", "mozilla", "konqueror", NULL};	//browsers to try
    gchar *argv[] = {NULL, PREFIX"/share/doc/z80asmide/manual.html", NULL};			//arg vector

    if (g_getenv("BROWSER") != NULL) {				//first, if there's a BROWSER var, try that
		argv[0] = g_getenv("BROWSER");
		if (g_spawn_async(NULL, argv, NULL, G_SPAWN_SEARCH_PATH, NULL, NULL, NULL, NULL))
	    	return;
    }
    
    int i;
    for (i = 0; browsers[i] != NULL; i++) {			//then try each one on the list
        argv[0] = browsers[i];
        if (g_spawn_async(NULL, argv, NULL, G_SPAWN_SEARCH_PATH, NULL, NULL, NULL, NULL))	//if it works, return
	    	return;
    }

    GtkWidget *dialog = gtk_message_dialog_new(NULL, GTK_DIALOG_NO_SEPARATOR, GTK_MESSAGE_ERROR, GTK_BUTTONS_OK, "A web browser could not be found on\nyour system.  You can view the manual yourself\nby opening "PREFIX"/share/doc/z80asmide/manual.html.");
    gtk_dialog_run(dialog);						//otherwise, show an error message
    gtk_widget_destroy(dialog);
}
