/*
    Z80 Assembly IDE - An IDE for the Z80 line of TI calculators
    Copyright (C) 2006 Donald Straney

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <gtk/gtk.h>
#include <string.h>

#include "asmide.h"
#include "settings.h"

/*
    Loads settings from file
    in user's home directory,
    sets properties
*/

void load_settings(struct status *curr_status) {

    GtkWindow *window = curr_status->window;
    gchar **color_label = curr_status->color_label;
    gchar **color_comment = curr_status->color_comment;
    gchar **color_directive = curr_status->color_directive;
    gchar **output_type = curr_status->output_type;
    gboolean *usesyncolor = curr_status->usesyncolor;
    gboolean *usewordwrap = curr_status->usewordwrap;
    gboolean *useautoindent = curr_status->useautoindent;
    gint *tabwidth = curr_status->tabwidth;
	PangoFontDescription **font = curr_status->font;
    gchar **incdir1 = curr_status->incdir1;
    gchar **incdir2 = curr_status->incdir2;
    gchar **incdir3 = curr_status->incdir3;

    gchar *path = g_build_filename(g_getenv("HOME"), ".z80asmide", NULL);		//get the path to the config file
    
    FILE *configfile = fopen(path, "r");	//and try to open the file
    
    if (configfile == NULL) {			//if it doesn't exist,
    
		gtk_window_set_default_size(window, DEFAULT_WINDOW_WIDTH, DEFAULT_WINDOW_HEIGHT);	//and use default settings
		*font = pango_font_description_from_string(DEFAULT_FONT);
		*color_label = g_strdup(DEFAULT_COLOR_LABEL);
		*color_comment = g_strdup(DEFAULT_COLOR_COMMENT);
		*color_directive = g_strdup(DEFAULT_COLOR_DIRECTIVE);
		*tabwidth = DEFAULT_TAB_WIDTH;
		*output_type = g_strdup(DEFAULT_OUTPUT_TYPE);
		*usesyncolor = TRUE;
		*usewordwrap = FALSE;
		*useautoindent = TRUE;

    } else {

		char rawinput[100], fontname[50];
		gint window_width = 0, window_height = 0;
		*usesyncolor = TRUE;
		*usewordwrap = FALSE;
		*useautoindent = TRUE;
		*tabwidth = 0;
		*font = pango_font_description_new();
	
		while (!feof(configfile) && fscanf(configfile, "%99[^\n]\n", &rawinput) != 0) {		//otherwise, read a line from the file

	    	if (g_str_has_prefix(rawinput, "WindowWidth="))			//parse recognized properties
				sscanf(rawinput + 12, "%d", &window_width);
		    else if (g_str_has_prefix(rawinput, "WindowHeight="))
				sscanf(rawinput + 13, "%d", &window_height);
	    	else if (g_str_has_prefix(rawinput, "FontFamily=")) {
				sscanf(rawinput + 11, "%49[^\n]\n", &fontname);
				pango_font_description_set_family(*font, fontname);
		    } else if (g_str_has_prefix(rawinput, "FontStyle=")) {
				int style;
				sscanf(rawinput + 10, "%d", &style);
				pango_font_description_set_style(*font, (PangoStyle)style);
			} else if (g_str_has_prefix(rawinput, "FontWeight=")) {
				int weight;
				sscanf(rawinput + 11, "%d", &weight);
				pango_font_description_set_weight(*font, (PangoWeight)weight);
			} else if (g_str_has_prefix(rawinput, "FontSize=")) {
				int size;
				sscanf(rawinput + 9, "%d", &size);
				pango_font_description_set_size(*font, size);
			} else if (g_str_has_prefix(rawinput, "ColorLabel="))
				*color_label = g_strndup(rawinput + 11, strlen(rawinput) - 11);
		    else if (g_str_has_prefix(rawinput, "ColorComment="))
				*color_comment = g_strndup(rawinput + 13, strlen(rawinput) - 13);
		    else if (g_str_has_prefix(rawinput, "ColorDirective="))
				*color_directive = g_strndup(rawinput + 15, strlen(rawinput) - 15);
		    else if (g_str_has_prefix(rawinput, "OutputType="))
				*output_type = g_strndup(rawinput + 11, strlen(rawinput) - 11);
		    else if (g_str_has_prefix(rawinput, "SyntaxColoring="))
				sscanf(rawinput + 15, "%d", usesyncolor);
		    else if (g_str_has_prefix(rawinput, "TabWidth="))
				sscanf(rawinput + 9, "%d", tabwidth);
		    else if (g_str_has_prefix(rawinput, "WordWrap="))
				sscanf(rawinput + 9, "%d", usewordwrap);
		    else if (g_str_has_prefix(rawinput, "IncDir1="))
				*incdir1 = g_strndup(rawinput + 8, strlen(rawinput) - 8);
		    else if (g_str_has_prefix(rawinput, "IncDir2="))
				*incdir2 = g_strndup(rawinput + 8, strlen(rawinput) - 8);
		    else if (g_str_has_prefix(rawinput, "IncDir3="))
				*incdir3 = g_strndup(rawinput + 8, strlen(rawinput) - 8);
		    else if (g_str_has_prefix(rawinput, "AutoIndent="))
				sscanf(rawinput + 11, "%d", useautoindent);
		    else
				fprintf(stderr, "z80asmide: Unrecognized property %s in config file\n", rawinput);		//if it's an invalid property, show an error
		}										//continue until the end of the file

		fclose(configfile);

		if (window_width <= 0)								//do some sanity checks for properties
	    	window_width = DEFAULT_WINDOW_WIDTH;
		if (window_height <= 0)
		    window_height = DEFAULT_WINDOW_HEIGHT;

		if (*color_label == NULL)
		    *color_label = g_strdup(DEFAULT_COLOR_LABEL);
		if (*color_comment == NULL)
		    *color_comment = g_strdup(DEFAULT_COLOR_COMMENT);
		if (*color_directive == NULL)
		    *color_directive = g_strdup(DEFAULT_COLOR_DIRECTIVE);

		if (*output_type == NULL)
		    *output_type = g_strdup(DEFAULT_OUTPUT_TYPE);

		if (*tabwidth <= 0)
		    *tabwidth = DEFAULT_TAB_WIDTH;

		gtk_window_set_default_size(window, window_width, window_height);		//then set the properties
	}
    
    g_free(path);
}


/*
    Saves settings to
    file in user's home
    directory
*/

void save_settings(struct status *curr_status) {

    GtkWindow *window = curr_status->window;
    gchar *color_label = *(curr_status->color_label);
    gchar *color_comment = *(curr_status->color_comment);
    gchar *color_directive = *(curr_status->color_directive);
    gchar *output_type = *(curr_status->output_type);
    gboolean usesyncolor = *(curr_status->usesyncolor);
    gboolean usewordwrap = *(curr_status->usewordwrap);
    gboolean useautoindent = *(curr_status->useautoindent);
    gint tabwidth = *(curr_status->tabwidth);
	PangoFontDescription *font = *(curr_status->font);
    gchar *incdir1 = *(curr_status->incdir1);
    gchar *incdir2 = *(curr_status->incdir2);
    gchar *incdir3 = *(curr_status->incdir3);

    gchar *path = g_build_filename(g_getenv("HOME"), ".z80asmide", NULL);		//get the path to the config file
    
    FILE *configfile = fopen(path, "w");	//and try to open the file

    if (configfile == NULL) {			//if it couldn't be opened,

		GtkWidget *dialog = gtk_message_dialog_new(NULL, 0, GTK_MESSAGE_WARNING, GTK_BUTTONS_OK, "Can't open configuration file\n%s to write settings!", path);
		gtk_message_dialog_format_secondary_text(dialog, "Your permissions must be really\nscrewed up if you don't have write\naccess to your home directory.");
		gtk_dialog_run(dialog);			//show a warning dialog
		gtk_widget_destroy(dialog);

    } else {				//otherwise,

		gint window_width, window_height;
		const char *fontname = pango_font_description_get_family(font);
		int fontstyle = (int)pango_font_description_get_style(font);
		int fontweight = (int)pango_font_description_get_weight(font);
		gint fontsize = pango_font_description_get_size(font);	//(shouldn't have to worry about different scales - this should always be in points because the font is set from the config file in points and the default font size is specified in points)
		gtk_window_get_size(window, &window_width, &window_height);	//get the properties
		fprintf(configfile, "WindowWidth=%d\nWindowHeight=%d\nFontFamily=%s\nFontWeight=%d\nFontStyle=%d\nFontSize=%d\nColorLabel=%s\nColorComment=%s\nColorDirective=%s\nOutputType=%s\nSyntaxColoring=%d\nWordWrap=%d\nAutoIndent=%d\nTabWidth=%d\n", window_width, window_height, fontname, fontweight, fontstyle, fontsize, color_label, color_comment, color_directive, output_type, usesyncolor, usewordwrap, useautoindent, tabwidth);	//and write them to the file
		if (incdir1 != NULL)
		    fprintf(configfile, "IncDir1=%s\n", incdir1);
		if (incdir2 != NULL)
		    fprintf(configfile, "IncDir2=%s\n", incdir2);
		if (incdir3 != NULL)
		    fprintf(configfile, "IncDir3=%s\n", incdir3);
		fclose(configfile);
    }

    g_free(path);
}
